!!****if* source/Simulation/SimulationMain/unitTest/ProtonImaging/notUsedFiles/sim_setupCrossInSquareXY
!!
!! NAME 
!!
!!  sim_setupCrossInSquareXY
!!
!! SYNOPSIS
!!
!!  sim_setupCrossInSquareXY ()
!!
!! DESCRIPTION
!!
!!  This routine sets up beam proton (x,y) coordinates, such that a cross in a square
!!  picture will result on the beam's crossectional plane. The picture looks like:
!!
!!                           1------------2
!!                           |\          /|
!!                           | \        / |       (x,y) of point 1 -> (-0.5,+0.5)
!!                           |  \      /  |
!!                           |   \    /   |       (x,y) of point 2 -> (+0.5,+0.5)
!!                           |    \  /    |
!!                           |     \/     |       (x,y) of point 3 -> (-0.5,-0.5)
!!                           |     /\     |
!!                           |    /  \    |       (x,y) of point 4 -> (+0.5,-0.5)
!!                           |   /    \   |
!!                           |  /      \  |       (x,y) of cross center -> (0,0)
!!                           | /        \ |
!!                           |/          \|
!!                           3------------4
!!
!!  These (x,y) beam crossectional coordinates are in units of the circular radial
!!  value, i.e. the radius of the circular beam is equal to 1.
!!
!!  On each of the square sides we place 1000 points. On each of the diagonals we
!!  place 1412 points (excluding the corner and center points). This maintains roughly
!!  equal distances between each consecutive points on the lines. The total number of
!!  points created will be: 4 * 1000 + 2 * 1412 + 1 = 6825
!!
!! ARGUMENTS
!!
!!***

subroutine sim_setupCrossInSquareXY ()

  use Driver_interface, ONLY : Driver_abortFlash

  use Simulation_Data,  ONLY : sim_nBeamPoints, &
                               sim_beamPointsX, &
                               sim_beamPointsY

  implicit none

  integer :: i,n
  real    :: s,x,y

  n = 0
!
!
!     ...The 1st side (x = from -0.499 to +0.500 in steps of 0.001, y = +0.500).
!
!
  x = -0.5
  y =  0.5

  do i = 1,1000
     n = n + 1
     x = x + 0.001
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The 2nd side (y = from +0.499 to -0.500 in steps of 0.001, x = +0.500).
!
!
  x = 0.5
  y = 0.5

  do i = 1,1000
     n = n + 1
     y = y - 0.001
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The 3rd side (x = from +0.499 to -0.500 in steps of 0.001, y = -0.500).
!
!
  x =  0.5
  y = -0.5

  do i = 1,1000
     n = n + 1
     x = x - 0.001
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The 4th side (y = from -0.499 to +0.500 in steps of 0.001, x = -0.500).
!
!
  x = -0.5
  y = -0.5

  do i = 1,1000
     n = n + 1
     y = y + 0.001
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...Calculate stepping size for diagonals.
!
!
  s = 0.5 / 706.0
!
!
!     ...The 1st diagonal half (x = - y).
!
!
  x = -0.5
  y =  0.5

  do i = 1,706
     n = n + 1
     x = x + s
     y = y - s
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The 2nd diagonal half (x = y).
!
!
  x = 0.5
  y = 0.5

  do i = 1,706
     n = n + 1
     x = x - s
     y = y - s
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The 3rd diagonal half (x = - y).
!
!
  x =  0.5
  y = -0.5

  do i = 1,706
     n = n + 1
     x = x - s
     y = y + s
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The 4th diagonal half (x = y).
!
!
  x = -0.5
  y = -0.5

  do i = 1,706
     n = n + 1
     x = x + s
     y = y + s
     sim_beamPointsX (n) = x
     sim_beamPointsY (n) = y
  end do
!
!
!     ...The center point.
!
!
  n = n + 1
  sim_beamPointsX (n) = 0.0
  sim_beamPointsY (n) = 0.0
!
!
!     ...Check.
!
!
  if (n /= sim_nBeamPoints) then
      call Driver_abortFlash ('[sim_setupCrossInSquareXY] ERROR: mismatch # of (x,y)-pairs created')
  end if
!
!
!     ...Ready!
!
!
  return
end subroutine sim_setupCrossInSquareXY
