!!****if* source/Simulation/SimulationMain/unitTest/ProtonImaging/pi_beam3DGridPointsStatistical
!!
!! NAME
!!
!!  pi_beam3DGridPointsStatistical
!!
!! SYNOPSIS
!!
!!  call pi_beam3DGridPointsStatistical (real,    intent (in)    :: radius,
!!                                       real,    intent (in)    :: seed,
!!                                       integer, intent (in)    :: targetTotalGridPoints,
!!                                       logical, intent (inout) :: startGrid,
!!                                       integer, intent (in)    :: maxGridPoints,
!!                                       logical, intent (out)   :: moreGridPoints,
!!                                       integer, intent (out)   :: nGridPoints,
!!                                       real,    intent (out)   :: xGrid (1:maxGridPoints),
!!                                       real,    intent (out)   :: yGrid (1:maxGridPoints))
!!
!! DESCRIPTION
!!
!!  Overriding version of the original. This particular version returns chuncks of (x,y) grid
!!  points corresponding to a particular grid point setup used for the current simulation.
!!  As many (x,y) pairs are returned as either fit in the memory or are still needed for the
!!  simulation setup.
!!
!! ARGUMENTS
!!
!!  radius                : the radius of the circular cross section of the 3D beam
!!  seed                  : irrelevant here
!!  targetTotalGridPoints : the complete number of grid points wanted
!!  startGrid             : if true, the local grid point count starts from 0
!!  maxGridPoints         : the maximum number of grid points that can be returned
!!  moreGridPoints        : if true, more grid points are expected
!!  nGridPoints           : the actual number of grid points returned
!!  xGrid                 : the (abstract) x-coordinates of the grid points
!!  yGrid                 : the (abstract) y-coordinates of the grid points
!!
!! NOTES
!!
!!  The stored (x,y) pairs corresponding to the simulation setup are in units of the radius
!!  of the beam, hence they have to be multiplied by the radius value.
!!
!!***

subroutine pi_beam3DGridPointsStatistical (radius,                             &
                                           seed,                               &
                                           targetTotalGridPoints,              &
                                           startGrid,                          &
                                           maxGridPoints,                      &
                                                               moreGridPoints, &
                                                               nGridPoints,    &
                                                               xGrid,          &
                                                               yGrid           )

  use Driver_interface,   ONLY : Driver_abortFlash

  use Simulation_data,    ONLY : sim_nBeamPoints, &
                                 sim_beamPointsX, &
                                 sim_beamPointsY

  implicit none

  real,    intent (in)    :: radius
  integer, intent (in)    :: seed
  integer, intent (in)    :: targetTotalGridPoints
  logical, intent (inout) :: startGrid
  integer, intent (in)    :: maxGridPoints
  logical, intent (out)   :: moreGridPoints
  integer, intent (out)   :: nGridPoints
  real,    intent (out)   :: xGrid (1:maxGridPoints)
  real,    intent (out)   :: yGrid (1:maxGridPoints)

  logical       :: fullGrid

  integer, save :: collectedBeamGridPoints = 0    ! save for next call, initial value is zero
  integer, save :: collectedLocalGridPoints       ! save for next call
  integer       :: n

  real          :: x, y
!
!
!     ...Start local grid point count if necessary.
!
!
  if (startGrid) then
      collectedLocalGridPoints = 0
      startGrid = .false.
  end if
!
!
!     ...Retrieve the current array of grid points.
!
!
  nGridPoints = 0

  do n = 1,maxGridPoints

     collectedBeamGridPoints  = collectedBeamGridPoints  + 1    ! global beam grid counter
     collectedLocalGridPoints = collectedLocalGridPoints + 1    ! local beam grid counter

     x = sim_beamPointsX (collectedBeamGridPoints)
     y = sim_beamPointsY (collectedBeamGridPoints)

     nGridPoints = nGridPoints + 1                              ! current grid counter

     xGrid (nGridPoints) = x
     yGrid (nGridPoints) = y

     fullGrid =     (collectedLocalGridPoints == targetTotalGridPoints) &
                .or.(collectedBeamGridPoints  == sim_nBeamPoints)

     if (fullGrid) then

         xGrid (1:nGridPoints) = radius * xGrid (1:nGridPoints)   ! convert to unit vector units
         yGrid (1:nGridPoints) = radius * yGrid (1:nGridPoints)   ! convert to unit vector units

         moreGridPoints = .false.
         return
     end if

  end do
!
!
!     ...Return the collected grid points. If this part of the code is reached, there
!        are more grid points to be collected.
!
!
  xGrid (1:nGridPoints) = radius * xGrid (1:nGridPoints)
  yGrid (1:nGridPoints) = radius * yGrid (1:nGridPoints)

  moreGridPoints = .true.
!
!
!     ...Ready!
!
!
  return
end subroutine pi_beam3DGridPointsStatistical
