!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonBeams/pi_beamsCheck2DCyl3D
!!
!! NAME
!!
!!  pi_beamsCheck2DCyl3D
!!
!! SYNOPSIS
!!
!!  call pi_beamsCheck2DCyl3D ()
!!
!! DESCRIPTION
!!
!!  Checks the collected 3D cartesian beams data for 2D cylindrical grids (cartesian). This is
!!  a rather specialized routine, in which the dimensions where the beams are defined is different
!!  from the dimension of the underlaying grid. All checks which depend on domain grid details
!!  should go in here. Currently it contains the following:
!!
!!         1) Check, if all beam circular lens areas are completely outside the 3D cylindrical
!!            domain, defined by the 2D cylindrical grid.
!!
!!  The 3D cylindrical domain is obtained from the (x=R,y=z) 2D cylindrical grid by a 360
!!  degree revolution around the 2D cylindrical z-axis.
!!
!!  Since the target area of the proton beam is merely used to construct the direction of
!!  the individual protons, there is no need to check if this area is properly located wrt
!!  to the 3D cylindrical domain.
!!
!! ARGUMENTS
!!
!! NOTES
!!
!!***

subroutine pi_beamsCheck2DCyl3D ()

  use Driver_interface,      ONLY : Driver_abortFlash

  use Roots_interface,       ONLY : Roots_x4Polynomial

  use ProtonImaging_data,    ONLY : pi_beams,               &
                                    pi_numberOfBeams,       &
                                    pi_orthogonalTolerance, &
                                    pi_xminDomain,          &
                                    pi_xmaxDomain,          &
                                    pi_yminDomain,          &
                                    pi_ymaxDomain
  
  implicit none

#include "Flash.h"
#include "ProtonImaging.h"
#include "constants.h"

  logical :: circular
  logical :: CisZero, DisZero
  logical :: inDomain
  logical :: useCosine

  integer :: beam
  integer :: n
  integer :: nReal
  integer :: nTrig

  integer, parameter :: Re = 1

  real    :: A, B, C, D
  real    :: aa, bb, cc, dd, ac, ad, bc, bd
  real    :: cosine, sine
  real    :: Cx1, Cx2, Cy1, Cy2, cz
  real    :: LdotL, r1dotL, r2dotL
  real    :: lensRadius
  real    :: lensRmin, lensZmin, lensZmax
  real    :: lensX, lensY, lensZ
  real    :: minRsq
  real    :: omega
  real    :: q0, q1, q2, q3, q3c, q3s, q4, q4inv
  real    :: r1dotr1, r1dotr2, r2dotr2
  real    :: r1x, r2x, r1y, r2y
  real    :: root1, root2, root3, root4
  real    :: Rsq1, Rsq2
  real    :: u1x, u1y, u1z
  real    :: u2x, u2y, u2z

  real    :: Trig (1:4)

  real    :: root (1:4,1:2)
!
!
!     ...Loop over all beams.
!
!
  do beam = 1, pi_numberOfBeams

     u1x        = pi_beams (beam) % axisUnit1X
     u1y        = pi_beams (beam) % axisUnit1Y
     u1z        = pi_beams (beam) % axisUnit1Z
     u2x        = pi_beams (beam) % axisUnit2X
     u2y        = pi_beams (beam) % axisUnit2Y
     u2z        = pi_beams (beam) % axisUnit2Z
     lensRadius = pi_beams (beam) % lensRadius
     lensX      = pi_beams (beam) % lensX
     lensY      = pi_beams (beam) % lensY
     lensZ      = pi_beams (beam) % lensZ
!
!
!     ...The (non-radial) z-component of the elliptical lens area is treated first.
!        Calculate the minimum and maximum lens area extension in z-direction.
!
!
     cz = lensRadius * sqrt (u1z * u1z + u2z * u2z)

     lensZmin = lensZ - cz
     lensZmax = lensZ + cz
!
!
!     ...Handle the (x,y) part of the circular lens area. Form the 2D projections onto
!        the (x,y) plane and get the relevant info.
!
!
     r1x = lensRadius * u1x
     r1y = lensRadius * u1y
     r2x = lensRadius * u2x
     r2y = lensRadius * u2y

     LdotL   = lensX * lensX + lensY * lensY
     r1dotL  = lensX * r1x   + lensY * r1y
     r2dotL  = lensX * r2x   + lensY * r2y
     r1dotr1 =   r1x * r1x   +   r1y * r1y
     r1dotr2 =   r1x * r2x   +   r1y * r2y
     r2dotr2 =   r2x * r2x   +   r2y * r2y

     C = r1dotr1 - r2dotr2
     D = r1dotr2

     CisZero = (abs (C) / max (r1dotr1, r2dotr2)) < pi_orthogonalTolerance
     DisZero = (abs (D) / max (r1dotr1, r2dotr2)) < pi_orthogonalTolerance

     circular = DisZero .and. CisZero
!
!
!     ...Set up the quartic coefficients and solve the appropriate quartic. Handle separately
!        the case when the projected 2D lens area is a circle. For the lens area we are
!        only interested in knowing if it is outside of the domain, so only the minimum
!        radial value is of interest.
!
!
     if (circular) then

         lensRmin = sqrt (LdotL) - sqrt (r1dotr1)   ! subtract lengths of projected r1 from projected L

     else

         A = r1dotL
         B = r2dotL

         aa = A * A
         bb = B * B
         cc = C * C
         dd = D * D
         ac = A * C
         ad = A * D
         bc = B * C
         bd = B * D

         q4    = cc + dd + dd + dd + dd
         q4inv = 1.0 / q4
         q3c   = bd + bd + bd + bd + ac + ac
         q3s   = ad + ad + ad + ad - bc - bc

         useCosine = abs (q3c) > abs (q3s)

         if (useCosine) then
             q3 = q3c * q4inv
             q2 = (aa + bb) * q4inv - 1.0
             q1 = - (ac + ac + bd + bd) * q4inv
             q0 = (dd - aa) * q4inv
         else
             q3 = q3s * q4inv
             q2 = (aa + bb) * q4inv - 1.0
             q1 = (bc + bc - ad - ad) * q4inv
             q0 = (dd - bb) * q4inv
         end if

         call Roots_x4Polynomial (q3,q2,q1,q0,   nReal,root)

         nTrig = 0

         root1 = root (1,Re)
         root2 = root (2,Re)
         root3 = root (3,Re)
         root4 = root (4,Re)

         if (nReal > 0 .and. abs (root1) <= 1.0) then
             nTrig = nTrig + 1
             Trig (nTrig) = root1
         end if

         if (nReal > 1 .and. abs (root2) <= 1.0) then 
             nTrig = nTrig + 1
             Trig (nTrig) = root2
         end if

         if (nReal > 2 .and. abs (root3) <= 1.0) then 
             nTrig = nTrig + 1
             Trig (nTrig) = root3
         end if

         if (nReal > 3 .and. abs (root4) <= 1.0) then 
             nTrig = nTrig + 1
             Trig (nTrig) = root4
         end if

         if (nTrig == 0) then
             call Driver_abortFlash ("pi_beamsCheck2DCyl3D: Quartic root failure for lens!")
         end if

         do n = 1,nTrig

            if (useCosine) then

                cosine = Trig (n)
                omega  = acos (cosine)
                sine   =  sin (omega)

                Cx1 = lensX + r1x * cosine + r2x * sine
                Cy1 = lensY + r1y * cosine + r2y * sine
                Cx2 = lensX + r1x * cosine - r2x * sine
                Cy2 = lensY + r1y * cosine - r2y * sine

            else

                sine   = Trig (n)
                omega  = asin (sine)
                cosine =  cos (omega)

                Cx1 = lensX + r1x * cosine + r2x * sine
                Cy1 = lensY + r1y * cosine + r2y * sine
                Cx2 = lensX - r1x * cosine + r2x * sine
                Cy2 = lensY - r1y * cosine + r2y * sine

            end if

            Rsq1 = Cx1 * Cx1 + Cy1 * Cy1
            Rsq2 = Cx2 * Cx2 + Cy2 * Cy2

            if (n == 1) then
                minRsq = min (Rsq1, Rsq2)
            else
                minRsq = min (minRsq , Rsq1, Rsq2)
            end if

         end do

         lensRmin = sqrt (minRsq)

     end if
!
!
!     ...Check, if beam lens area touches the domain.
!
!
     inDomain =      (lensZmin >= pi_yminDomain) &
               .and. (lensZmax <= pi_ymaxDomain) &
               .and. (lensRmin <= pi_xmaxDomain)

     if (inDomain) then
         call Driver_abortFlash ("pi_beamsCheck2DCyl3D: Beam lens (partially) inside the domain!")
     end if

  end do       ! loop over beams
!
!
!     ...Ready!
!
!
  return
end subroutine pi_beamsCheck2DCyl3D
