!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonBeams/pi_beamsCheck3DRec
!!
!! NAME
!!
!!  pi_beamsCheck3DRec
!!
!! SYNOPSIS
!!
!!  call pi_beamsCheck3DRec ()
!!
!! DESCRIPTION
!!
!!  Checks the collected beams data for those geometries consisting formally of 3D rectangular
!!  grids (cartesian). All checks which depend on domain grid details should go in here.
!!  Currently it contains the following:
!!
!!         1) Check, if all beam circular lens areas are completely outside the domain.
!!
!!  Since the target area of the proton beam is merely used to construct the direction of
!!  the individual protons, there is no need to check if this area is properly located wrt
!!  to the domain.
!!
!! ARGUMENTS
!!
!! NOTES
!!
!!***

subroutine pi_beamsCheck3DRec ()

  use Driver_interface,      ONLY : Driver_abortFlash

  use ProtonImaging_data,    ONLY : pi_beams,         &
                                    pi_numberOfBeams, &
                                    pi_xminDomain,    &
                                    pi_xmaxDomain,    &
                                    pi_yminDomain,    &
                                    pi_ymaxDomain,    &
                                    pi_zminDomain,    &
                                    pi_zmaxDomain
  
  implicit none

#include "Flash.h"
#include "ProtonImaging.h"
#include "constants.h"

  logical :: inDomain

  integer :: beam

  real    :: cx, cy, cz
  real    :: lensRadius
  real    :: lensX, lensY, lensZ
  real    :: u1x, u1y, u1z
  real    :: u2x, u2y, u2z
!
!
!     ...Loop over all beams.
!
!
  do beam = 1, pi_numberOfBeams

     u1x        = pi_beams (beam) % axisUnit1X
     u1y        = pi_beams (beam) % axisUnit1Y
     u1z        = pi_beams (beam) % axisUnit1Z
     u2x        = pi_beams (beam) % axisUnit2X
     u2y        = pi_beams (beam) % axisUnit2Y
     u2z        = pi_beams (beam) % axisUnit2Z
     lensRadius = pi_beams (beam) % lensRadius
     lensX      = pi_beams (beam) % lensX
     lensY      = pi_beams (beam) % lensY
     lensZ      = pi_beams (beam) % lensZ
!
!
!     ...The implicit form of the circular lens boundary curve is:
!
!                c(t) = |r| * [u1 * cos (t)  +  u2 * sin (t)]
!
!        where u1,u2 stand for the two orthogonal unit vectors in the local lens coordinate
!        system, |r| is the radius of the circular lens area, 'c(t)' is a vector on the circular
!        boundary and 't' is the implicit parameter ranging from 0 to 2pi. Differentiating this
!        equation with respect to 't' we obtain:
!
!                     t = arctan ( u2 / u1 )
!
!        which, when inserted into c(t), leads to the simple minimax equation:
!
!                  c (min,max) = +/- |r| * sqrt ([u1]^2 + [u2]^2)
!
!        or, in component form:
!
!                 cx (min,max) = +/- |r| * sqrt ([u1x]^2 + [u2x]^2)
!                 cy (min,max) = +/- |r| * sqrt ([u1y]^2 + [u2y]^2)
!                 cz (min,max) = +/- |r| * sqrt ([u1z]^2 + [u2z]^2)
!
!        Since the domain boundaries are in terms of the global coordinate system, we need
!        to convert the circular curve points from the local to the global coordinate
!        system via:
!
!                                C = c + L
!
!        where L is the position of the lens center. The results of C will then
!        be tested against the domain boundaries.
!
!
     cx = lensRadius * sqrt (u1x * u1x + u2x * u2x)
     cy = lensRadius * sqrt (u1y * u1y + u2y * u2y)
     cz = lensRadius * sqrt (u1z * u1z + u2z * u2z)
!
!
!     ...Check, if beam lens area touches the domain.
!
!
     inDomain =      (lensX - cx >= pi_xminDomain) &
               .and. (lensY - cy >= pi_yminDomain) &
               .and. (lensZ - cz >= pi_zminDomain) &
               .and. (lensX + cx <= pi_xmaxDomain) &
               .and. (lensY + cy <= pi_ymaxDomain) &
               .and. (lensZ + cz <= pi_zmaxDomain)

     if (inDomain) then
         call Driver_abortFlash ("pi_beamsCheck3DRec: Beam lens (partially) inside the domain!")
     end if

  end do
!
!
!     ...Ready!
!
!
  return
end subroutine pi_beamsCheck3DRec
