!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonBeams/pi_beamsInfo3DRec
!!
!! NAME
!!
!!  pi_beamsInfo3DRec
!!
!! SYNOPSIS
!!
!!  call pi_beamsInfo3DRec ()
!!
!! DESCRIPTION
!!
!!  Generates information about the proton beams for those domain geometries consisting
!!  formally of 3D rectangular grids (cartesian). In here all beam information is generated
!!  that can be generated at initialization. Currently it contains the following:
!!
!!         1) Convert proton energy from MeV to initial proton speed in units
!!            of cm/s.
!!
!!         2) Calculate the lens to target distance.
!!
!!         3) Calculate basic information about the planar circular beam proton grid.
!!
!!         4) Calculate two orthogonal unit vectors within the circular plane of
!!            of the lens located at the lens' center.
!!
!! ARGUMENTS
!!
!!  none
!!
!! NOTES
!!
!!***

subroutine pi_beamsInfo3DRec ()

  use Driver_interface,    ONLY : Driver_abortFlash

  use Logfile_interface,   ONLY : Logfile_stamp

  use pi_interface,        ONLY : pi_beam3DGridSetupStatistical

  use ProtonImaging_data,  ONLY : pi_beams,                 &
                                  pi_MeV2erg,               &
                                  pi_normalizedTolerance,   &
                                  pi_numberOfBeams,         &
                                  pi_orthogonalTolerance,   &
                                  pi_protonMass,            &
                                  pi_speedOfLight,          &
                                  pi_speedOfLightSquared

  implicit none

#include "ProtonImaging.h"
#include "Flash.h"
#include "constants.h"

  logical :: planeXY, planeXZ
  logical :: seedIncrement, seedInitialize

  integer :: beam
  integer :: seed, seedMaximum, seedStepping

  real    :: apertureAngle
  real    :: beamLength, beamLengthInv
  real    :: bProj, bProjInv
  real    :: bx, by, bz
  real    :: E, mc2, v
  real    :: lensRadius
  real    :: lensX, lensY, lensZ
  real    :: norm1, norm2
  real    :: ortho12, ortho1b, ortho2b
  real    :: protonEnergy
  real    :: target2LensMagnification
  real    :: targetRadius
  real    :: targetX, targetY, targetZ
  real    :: u1x, u1y, u1z
  real    :: u2x, u2y, u2z
!
!
!     ...Loop over all beams.
!
!
  do beam = 1, pi_numberOfBeams

     apertureAngle       = pi_beams (beam) % apertureAngle
     lensRadius          = pi_beams (beam) % lensRadius
     lensX               = pi_beams (beam) % lensX
     lensY               = pi_beams (beam) % lensY
     lensZ               = pi_beams (beam) % lensZ
     protonEnergy        = pi_beams (beam) % protonEnergy
     targetX             = pi_beams (beam) % targetX
     targetY             = pi_beams (beam) % targetY
     targetZ             = pi_beams (beam) % targetZ
!
!
!     ...Set the dimensionality indicator of the beam.
!
!
     pi_beams (beam) % dimensionality = 3
!
!
!     ...Calculate the initital proton speed (in units of cm/s) from the supplied
!        proton energy in MeV.
!
!
     E   = protonEnergy * pi_MeV2erg
     mc2 = pi_protonMass * pi_speedOfLightSquared
     v   = sqrt (1.0 - (mc2 / (E + mc2)) ** 2)     ! in units of light speed

     pi_beams (beam) % initialProtonSpeed = v * pi_speedOfLight
!
!
!     ...Calculate distance from lens to target and store in beams array.
!
!
     bx = lensX - targetX         ! beam vector pointing from target to lens
     by = lensY - targetY
     bz = lensZ - targetZ

     beamLength    = sqrt (bx * bx + by * by + bz * bz)
     beamLengthInv = 1.0 / beamLength

     pi_beams (beam) % distanceLens2Target = beamLength
!
!
!     ...Set up the beam's statistical grid and store the info.
!
!
     seedInitialize = .true.
     seedIncrement  = .false.

     call pi_beam3DGridSetupStatistical (seedInitialize,                &
                                         seedIncrement,                 &
                                                          seedMaximum,  &
                                                          seedStepping, &
                                                          seed          )

     pi_beams (beam) % gridSeed           = seed
     pi_beams (beam) % gridSeedMaximum    = seedMaximum
     pi_beams (beam) % gridSeedStepping   = seedStepping
!
!
!     ...Calculate and store the target radius length. Calculate also the magnification
!        factor for target -> lens. Since the lens is always smaller or equal in size
!        compared to the target, this magnification factor is <= 1. The magnification factor
!        will be needed to establish the beam grid data at the lens, since all the beam grid
!        data that will be stored is based on the target.
!
!
     targetRadius             = lensRadius + tan (apertureAngle * 0.5) * beamLength
     target2LensMagnification = lensRadius  / targetRadius

     pi_beams (beam) % targetRadius             = targetRadius
     pi_beams (beam) % target2LensMagnification = target2LensMagnification
!
!
!     ...Set up 2 orthogonal unit vectors in terms of the local lens/target x,y,z-coordinate
!        system, whose origin coincides with center location of the lens/target. The 1st
!        unit vector will be aligned with one of the global X,Y,Z axis, the choice of the
!        axis depending on the magnitude of the beam vector projection onto the XY,XZ,YZ
!        planes:
!
!               XY-plane projection of beam vector largest -> align along Z axis
!               XZ-plane projection of beam vector largest -> align along Y axis
!               YZ-plane projection of beam vector largest -> align along X axis
!
!               align along Z axis -> means: u1 torsion angle with Z axis is zero
!               align along Y axis -> means: u1 torsion angle with Y axis is zero
!               align along X axis -> means: u1 torsion angle with X axis is zero
!
!
     planeXY = (abs (bx) >= abs (by)) .and. (abs (by) >= abs (bz))      ! |bz| is smallest
     planeXZ = (abs (bx) >= abs (by)) .and. (abs (by) <  abs (bz))      ! |by| is smallest

     if (planeXY) then

         bProj    = sqrt (bx * bx + by * by)
         bProjInv = 1.0 / bProj

         u1x = - bx * bz * bProjInv * beamLengthInv
         u1y = - by * bz * bProjInv * beamLengthInv
         u1z =   bProj * beamLengthInv

         u2x = - by * bProjInv
         u2y =   bx * bProjInv
         u2z = 0.0

     else if (planeXZ) then

         bProj    = sqrt (bx * bx + bz * bz)
         bProjInv = 1.0 / bProj

         u1x = - bx * by * bProjInv * beamLengthInv
         u1y =   bProj * beamLengthInv
         u1z = - by * bz * bProjInv * beamLengthInv

         u2x =   bz * bProjInv
         u2y = 0.0
         u2z = - bx * bProjInv

     else

         bProj    = sqrt (by * by + bz * bz)
         bProjInv = 1.0 / bProj

         u1x =   bProj * beamLengthInv
         u1y = - bx * by * bProjInv * beamLengthInv
         u1z = - bx * bz * bProjInv * beamLengthInv

         u2x = 0.0
         u2y = - bz * bProjInv
         u2z =   by * bProjInv

     end if
!
!
!     ...unit vector components are ready. Check for unit vector conditions and if ok
!        store into beams array.
!
!
     norm1   = u1x * u1x  +  u1y * u1y  +  u1z * u1z
     norm2   = u2x * u2x  +  u2y * u2y  +  u2z * u2z
     ortho12 = u1x * u2x  +  u1y * u2y  +  u1z * u2z
     ortho1b = u1x *  bx  +  u1y *  by  +  u1z *  bz
     ortho2b = u2x *  bx  +  u2y *  by  +  u2z *  bz

     if (     (norm1 > 1.0 + pi_normalizedTolerance) &
         .or. (norm1 < 1.0 - pi_normalizedTolerance) ) then
          call Driver_abortFlash ("pi_beamsInfo3DRec: 1st unit vector not normalized!")
     end if

     if (     (norm2 > 1.0 + pi_normalizedTolerance) &
         .or. (norm2 < 1.0 - pi_normalizedTolerance) ) then
          call Driver_abortFlash ("pi_beamsInfo3DRec: 2nd unit vector not normalized!")
     end if

     if (     (ortho12 >   pi_orthogonalTolerance) &
         .or. (ortho12 < - pi_orthogonalTolerance) ) then
          call Driver_abortFlash ("pi_beamsInfo3DRec: 1st + 2nd unit vectors not orthogonal!")
     end if

     if (     (ortho1b >   pi_orthogonalTolerance) &
         .or. (ortho1b < - pi_orthogonalTolerance) ) then
          call Driver_abortFlash ("pi_beamsInfo3DRec: 1st unit + beam vectors not orthogonal!")
     end if

     if (     (ortho2b >   pi_orthogonalTolerance) &
         .or. (ortho2b < - pi_orthogonalTolerance) ) then
          call Driver_abortFlash ("pi_beamsInfo3DRec: 2nd unit + beam vectors not orthogonal!")
     end if

     pi_beams (beam) % axisUnit1X = u1x
     pi_beams (beam) % axisUnit1Y = u1y
     pi_beams (beam) % axisUnit1Z = u1z
     pi_beams (beam) % axisUnit2X = u2x
     pi_beams (beam) % axisUnit2Y = u2y
     pi_beams (beam) % axisUnit2Z = u2z

  end do
!
!
!     ...Ready!
!
!
  return
end subroutine pi_beamsInfo3DRec
