!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonBeams/save/pi_beamsCheck2DRec
!!
!! NAME
!!
!!  pi_beamsCheck2DRec
!!
!! SYNOPSIS
!!
!!  call pi_beamsCheck2DRec ()
!!
!! DESCRIPTION
!!
!!  Checks the collected beams data for those geometries consisting formally of 2D rectangular
!!  grids (cartesian + cylindrical). All checks which depend on domain grid details should go
!!  in here. Currently it contains the following:
!!
!!         1) Check, if all beam lens lines are completely outside the domain.
!!
!!  Since the target area of the proton beam is merely used to construct the direction of
!!  the individual protons, there is no need to check if this area is properly located wrt
!!  to the domain.
!!
!! ARGUMENTS
!!
!! NOTES
!!
!!***

subroutine pi_beamsCheck2 DRec ()

  use Driver_interface,      ONLY : Driver_abortFlash

  use ProtonImaging_data,    ONLY : pi_beams,         &
                                    pi_gridGeometry,  &
                                    pi_numberOfBeams, &
                                    pi_xminDomain,    &
                                    pi_xmaxDomain,    &
                                    pi_yminDomain,    &
                                    pi_ymaxDomain
  
  implicit none

#include "Flash.h"
#include "ProtonImaging.h"
#include "constants.h"

  logical :: inDomain

  integer :: beam

  real    :: lensRadius
  real    :: lensX, lensY
  real    :: lx, ly
  real    :: ux, uy
  real    :: xmaxDomain, xminDomain
  real    :: ymaxDomain, yminDomain
!
!
!     ...In case of 2D cylindrical geometry, treat the 2D cylindrical domain as the fully
!        3-dimensional rotational domain. This allows placing lenses and targets in the
!        negative x-coordinates area.
!
!
  xmaxDomain = pi_xmaxDomain
  yminDomain = pi_yminDomain
  ymaxDomain = pi_ymaxDomain

  if (pi_gridGeometry == GRID_2DCYLINDRICAL) then
      xminDomain = - pi_xmaxDomain
  else
      xminDomain =   pi_xminDomain
  end if
!
!
!     ...Loop over all beams.
!
!
  do beam = 1, pi_numberOfBeams

     ux         = pi_beams (beam) % axisUnit1X
     uy         = pi_beams (beam) % axisUnit1Y
     lensRadius = pi_beams (beam) % lensRadius
     lensX      = pi_beams (beam) % lensX
     lensY      = pi_beams (beam) % lensY
!
!
!     ...Any point on a line along the lens unit vectors is given by:
!
!                               l(t) = t * u
!
!        where 'u' stand for the line unit vector in the local lens coordinate system,
!        't' is a parameter from - infinity to + infinity and 'l(t)' is a vector on
!        the infinite line containing the lens line. The maximum values that this line
!        can attend for the lens are simply:
!
!                          l (min,max) = +/- r * u
!
!        where r is half the lens line length (radius). In component form we have:
!
!                         lx (min,max) = +/- r * ux
!                         ly (min,max) = +/- r * uy
!
!        Since the domain boundaries are in terms of the global coordinate system, we
!        need to convert the lens line points from the local to the global coordinate
!        system via forming 'l + L', where L is the position of the lens center.
!        The results of 'l + L' will then be tested against the domain boundaries.
!
!
     lx = lensRadius * ux
     ly = lensRadius * uy
!
!
!     ...Check, if beam lens line touches the domain.
!
!
     inDomain =      (lensX - lx >= xminDomain) &
               .and. (lensY - ly >= yminDomain) &
               .and. (lensX + lx <= xmaxDomain) &
               .and. (lensY + ly <= ymaxDomain)

     if (inDomain) then
         call Driver_abortFlash ("pi_beamsCheck2DRec: Beam lens (partially) inside the domain!")
     end if

  end do
!
!
!     ...Ready!
!
!
  return
end subroutine pi_beamsCheck2DRec
