!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonBeams/save/pi_beamsInfo2DRec
!!
!! NAME
!!
!!  pi_beamsInfo2DRec
!!
!! SYNOPSIS
!!
!!  call pi_beamsInfo2DRec ()
!!
!! DESCRIPTION
!!
!!  Generates information about the proton beams for those domain geometries consisting
!!  formally of 2D rectangular grids (cartesian + cylindrical). In here all beam information
!!  is generated that can be generated at initialization. Currently it contains the following:
!!
!!         1) Calculate the lens to target distance.
!!
!!         2) Calculate basic information about the planar circular beam proton grid.
!!
!!         3) Calculate two orthogonal unit vectors within the circular plane of
!!            of the lens/target located at the lens/target center.
!!
!! ARGUMENTS
!!
!!  none
!!
!! NOTES
!!
!!***

subroutine pi_beamsInfo2DRec ()

  use Driver_interface,    ONLY : Driver_abortFlash

  use Logfile_interface,   ONLY : Logfile_stamp

  use pi_interface,        ONLY : pi_beamGridSetupStatistical

  use ProtonImaging_data,  ONLY : pi_beams,                 &
                                  pi_normalizedTolerance,   &
                                  pi_numberOfBeams,         &
                                  pi_orthogonalTolerance

  implicit none

#include "EnergyDeposition.h"
#include "Flash.h"
#include "constants.h"

  logical :: planeXY, planeXZ
  logical :: seedIncrement, seedInitialize

  integer :: beam
  integer :: seed, seedMaximum, seedStepping

  real    :: apertureAngle
  real    :: beamLength, beamLengthInv
  real    :: bProj, bProjInv
  real    :: bx, by
  real    :: lensRadius
  real    :: lensX, lensY
  real    :: norm1, norm2
  real    :: ortho12, ortho1b, ortho2b
  real    :: target2LensMagnification
  real    :: targetRadius
  real    :: targetX, targetY
  real    :: ux, uy
!
!
!     ...Loop over all beams.
!
!
  do beam = 1, pi_numberOfBeams

     apertureAngle       = pi_beams (beam) % apertureAngle
     lensRadius          = pi_beams (beam) % lensRadius
     lensX               = pi_beams (beam) % lensX
     lensY               = pi_beams (beam) % lensY
     targetX             = pi_beams (beam) % targetX
     targetY             = pi_beams (beam) % targetY
!
!
!     ...Set the dimensionality indicator of the beam.
!
!
     pi_beams (beam) % dimensionality = 2
!
!
!     ...Calculate distance from lens to target and store in beams array.
!
!
     bx = lensX - targetX         ! beam vector pointing from target to lens
     by = lensY - targetY

     beamLength    = sqrt (bx * bx + by * by)
     beamLengthInv = 1.0 / beamLength

     pi_beams (beam) % distanceLens2Target = beamLength
!
!
!     ...Set up the beam's statistical grid and store the info.
!
!
     seedInitialize = .true.
     seedIncrement  = .false.

     call pi_beamGridSetupStatistical (seedInitialize,                &
                                       seedIncrement,                 &
                                                        seedMaximum,  &
                                                        seedStepping, &
                                                        seed          )

     pi_beams (beam) % gridSeed           = seed
     pi_beams (beam) % gridSeedMaximum    = seedMaximum
     pi_beams (beam) % gridSeedStepping   = seedStepping
!
!
!     ...Calculate and store the target radius length. Calculate also the magnification
!        factor for target -> lens. Since the lens is always smaller or equal in size
!        compared to the target, this magnification factor is <= 1. The magnification factor
!        will be needed to establish the beam grid data at the lens, since all the beam grid
!        data that will be stored is based on the target.
!
!
     targetRadius             = lensRadius + tan (apertureAngle * 0.5) * beamLength
     target2LensMagnification = lensRadius  / targetRadius

     pi_beams (beam) % targetRadius             = targetRadius
     pi_beams (beam) % target2LensMagnification = target2LensMagnification
!
!
!     ...Set up the unit vector in terms of the local lens/target x,y-coordinate
!        system, whose origin coincides with center location of the lens/target.
!
!
     ux = - by * beamLengthInv
     uy =   bx * beamLengthInv

     pi_beams (beam) % axisUnit1X = ux
     pi_beams (beam) % axisUnit1Y = uy

  end do
!
!
!     ...Ready!
!
!
  return
end subroutine pi_beamsInfo2DRec
