!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonDetection/pi_detectorsCheck3DRec
!!
!! NAME
!!
!!  pi_detectorsCheck3DRec
!!
!! SYNOPSIS
!!
!!  call pi_detectorsCheck3DRec ()
!!
!! DESCRIPTION
!!
!!  Checks the collected detectors data for those geometries consisting formally of 3D rectangular
!!  grids (cartesian). All checks which depend on domain grid details should go in here.
!!  Currently it contains the following:
!!
!!         1) Check, if all detector screen areas are completely outside of the domain.
!!
!! ARGUMENTS
!!
!! NOTES
!!
!!***

subroutine pi_detectorsCheck3DRec ()

  use Driver_interface,      ONLY : Driver_abortFlash

  use ProtonImaging_data,    ONLY : pi_detectors,         &
                                    pi_numberOfDetectors, &
                                    pi_xminDomain,        &
                                    pi_xmaxDomain,        &
                                    pi_yminDomain,        &
                                    pi_ymaxDomain,        &
                                    pi_zminDomain,        &
                                    pi_zmaxDomain
  
  implicit none

#include "Flash.h"
#include "ProtonImaging.h"
#include "constants.h"

  logical :: inDomain

  integer :: detector

  real    :: centerX, centerY, centerZ
  real    :: k1, k2, k3, k4
  real    :: kxmin, kxmax, kymin, kymax, kzmin, kzmax
  real    :: sideLength, sideLengthHalf
  real    :: uXx, uXy, uXz
  real    :: uYx, uYy, uYz
!
!
!     ...Loop over all detectors.
!
!
  do detector = 1, pi_numberOfDetectors

     centerX    = pi_detectors (detector) % centerX
     centerY    = pi_detectors (detector) % centerY
     centerZ    = pi_detectors (detector) % centerZ
     sideLength = pi_detectors (detector) % sideLength
     uXx        = pi_detectors (detector) % axisXunitX
     uXy        = pi_detectors (detector) % axisXunitY
     uXz        = pi_detectors (detector) % axisXunitZ
     uYx        = pi_detectors (detector) % axisYunitX
     uYy        = pi_detectors (detector) % axisYunitY
     uYz        = pi_detectors (detector) % axisYunitZ

     sideLengthHalf = 0.5 * sideLength
!
!
!     ...Since the detector screens are square, the extremum values in each direction
!        correspond to one of the four corners. Using the center C global X,Y,Z coordinates
!        and the local screen x,y,z coordinates, each corner K has the following global
!        coordinate:
!
!                   K(X,Y,Z) = C(X,Y,Z) + S/2 * [+/- uX(x,y,z) +/- uY(x,y,z)]
!
!        where s is the screen's side length and u1,u2 are the unit axis vectors defining
!        the screen area:
!
!
!                                  - K\
!                                  | | \
!                                  | |  \
!                                  | |   \
!                                  | | uY|\
!                                  S |   | \          C = detector center
!                                  | |   |  \
!                      detector    | |   |   K        K = screen corners
!                                  | |   C   |
!                       screen     - K    \  |        S = screen side length
!                                     \    \ |
!                                      \  uX\|        uX,uY = screen unit vectors
!                                       \    |
!                                        \   |
!                                         \  |
!                                          \ |
!                                           K|
!
!
!        For each X,Y,Z component of the corners we need to find the maximum and the minimum
!        between the four corner values and compare it to the domain extensions.
!
!
     k1 = centerX + sideLengthHalf * (   uXx + uYx)
     k2 = centerX + sideLengthHalf * (   uXx - uYx)
     k3 = centerX + sideLengthHalf * ( - uXx + uYx)
     k4 = centerX + sideLengthHalf * ( - uXx - uYx)

     kxmin = min (k1,k2,k3,k4)
     kxmax = max (k1,k2,k3,k4)

     k1 = centerY + sideLengthHalf * (   uXy + uYy)
     k2 = centerY + sideLengthHalf * (   uXy - uYy)
     k3 = centerY + sideLengthHalf * ( - uXy + uYy)
     k4 = centerY + sideLengthHalf * ( - uXy - uYy)

     kymin = min (k1,k2,k3,k4)
     kymax = max (k1,k2,k3,k4)

     k1 = centerZ + sideLengthHalf * (   uXz + uYz)
     k2 = centerZ + sideLengthHalf * (   uXz - uYz)
     k3 = centerZ + sideLengthHalf * ( - uXz + uYz)
     k4 = centerZ + sideLengthHalf * ( - uXz - uYz)

     kzmin = min (k1,k2,k3,k4)
     kzmax = max (k1,k2,k3,k4)
!
!
!     ...Check, if detector screen area touches the domain.
!
!
     inDomain =      (kxmin >= pi_xminDomain) &
               .and. (kymin >= pi_yminDomain) &
               .and. (kzmin >= pi_zminDomain) &
               .and. (kxmax <= pi_xmaxDomain) &
               .and. (kymax <= pi_ymaxDomain) &
               .and. (kzmax <= pi_zmaxDomain)

     if (inDomain) then
         call Driver_abortFlash ("pi_detectorsCheck3DRec: Detector screen (partially) inside the domain!")
     end if

  end do
!
!
!     ...Ready!
!
!
  return
end subroutine pi_detectorsCheck3DRec
