!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonDetection/pi_detectorsInfo3DRec
!!
!! NAME
!!
!!  pi_detectorsInfo3DRec
!!
!! SYNOPSIS
!!
!!  call pi_detectorsInfo3DRec ()
!!
!! DESCRIPTION
!!
!!  Generates information about the proton detectors placed in 3D rectangular (cartesian)
!!  space. In here all detector information is generated that can be generated at initialization.
!!  Currently it contains the following:
!!
!!         1) Calculate 1/2 and inverse of side length (for efficiency purposes).
!!
!!         2) Calculate the normal vector (if detector is aligned wrt to a proton beam).
!!
!!         3) Calculate the unit normal vector.
!!
!!         4) Calculate two orthogonal (x,y) axis unit vectors within the square screen
!!            plane of each detector, which will serve as a 2D cartesian local basis for
!!            the screen, located at the screen center.
!!
!! ARGUMENTS
!!
!!  none
!!
!! NOTES
!!
!!***

subroutine pi_detectorsInfo3DRec ()

  use Driver_interface,    ONLY : Driver_abortFlash

  use Logfile_interface,   ONLY : Logfile_stamp

  use ProtonImaging_data,  ONLY : pi_badTiltingAxis,        &
                                  pi_beams,                 &
                                  pi_detectors,             &
                                  pi_normalizedTolerance,   &
                                  pi_numberOfDetectors,     &
                                  pi_orthogonalTolerance

  implicit none

#include "ProtonImaging.h"
#include "Flash.h"
#include "constants.h"

  character (len = 1) :: sideTiltingAxis

  integer :: alignWRTbeamNr
  integer :: detector

  real    :: centerX, centerY, centerZ
  real    :: cosPhi, sinPhi
  real    :: distance2beamLens
  real    :: lensX, lensY, lensZ
  real    :: normX, normY
  real    :: normalLengthInv
  real    :: nProj, nProjInv
  real    :: nx, ny, nz, nxny, nxnz, nynz
  real    :: orthoXY, orthoXn, orthoYn
  real    :: sideLength
  real    :: sideTiltingAngle
  real    :: targetX, targetY, targetZ
  real    :: uXx, uXy, uXz
  real    :: uYx, uYy, uYz
!
!
!     ...Loop over all detectors.
!
!
  do detector = 1, pi_numberOfDetectors
!
!
!     ...Calculate 1/2 and inverse of the detectors side length. This avoids having to
!        recompute it for each proton during screen detection.
!
!
     sideLength = pi_detectors (detector) % sideLength

     pi_detectors (detector) % sideLengthHalf = 0.5 * sideLength
     pi_detectors (detector) % sideLengthInv  = 1.0 / sideLength
!
!
!     ...If the detector should be aligned wrt a beam, calculate the unit normal vector
!        and the center position of the detector screen. Otherwise form the unit normal
!        vector from the normal vector.
!
!
     alignWRTbeamNr = pi_detectors (detector) % alignWRTbeamNr

     if (alignWRTbeamNr > 0) then

         lensX   = pi_beams (alignWRTbeamNr) % lensX
         lensY   = pi_beams (alignWRTbeamNr) % lensY
         lensZ   = pi_beams (alignWRTbeamNr) % lensZ
         targetX = pi_beams (alignWRTbeamNr) % targetX
         targetY = pi_beams (alignWRTbeamNr) % targetY
         targetZ = pi_beams (alignWRTbeamNr) % targetZ

         nx = targetX - lensX      ! detector normal vector pointing from beam lens to target
         ny = targetY - lensY
         nz = targetZ - lensZ

         normalLengthInv = 1.0 / sqrt (nx * nx + ny * ny + nz * nz)

         nx = nx * normalLengthInv
         ny = ny * normalLengthInv
         nz = nz * normalLengthInv

         distance2beamLens = pi_detectors (detector) % distance2beamLens

         pi_detectors (detector) % centerX = lensX + nx * distance2beamLens
         pi_detectors (detector) % centerY = lensY + ny * distance2beamLens
         pi_detectors (detector) % centerZ = lensZ + nz * distance2beamLens

     else

         nx = pi_detectors (detector) % normalX
         ny = pi_detectors (detector) % normalY
         nz = pi_detectors (detector) % normalZ

         normalLengthInv = 1.0 / sqrt (nx * nx + ny * ny + nz * nz)

         nx = nx * normalLengthInv
         ny = ny * normalLengthInv
         nz = nz * normalLengthInv

     end if

     pi_detectors (detector) % normalX = nx
     pi_detectors (detector) % normalY = ny
     pi_detectors (detector) % normalZ = nz
!
!
!     ...Set up 2 orthogonal unit vectors uX and uY (the 2D detector screen basis vectors) in
!        terms of the local x,y,z-coordinate system, whose origin coincides with center location
!        of the detector screen:
!
!
!                                    |\
!                                    | \
!                                    |  \
!                                    |   \
!                                    | uY|\
!                                    |   | \                C = detector center
!                                    |   |  \
!                      detector      |   |   |
!                                    |   C---|---> n        n = normal unit vector
!                       screen       \    \  |
!                                     \    \ |
!                                      \  uX\|        uX,uY,n = form a right hand rule
!                                       \    |                  orthogonal vector system:
!                                        \   |                      n = uY x uX
!                                         \  |
!                                          \ |
!                                           \|
!
!
!        The Y unit vector uY will be aligned and tilted with respect to the tilting axis
!        (one of the global X,Y,Z axis). The tilting angle A is defined as follows: 1) initially
!        (tilting angle = 0) the three vectors uY, n and the tilting axis are coplanar, 2) the
!        tilting angle is the clockwise angle rotation of uY from the tilting axis along the
!        normal vector n:
!
!
!                                 tilting
!                                  axis
!                                    |
!                                    |     uY           This view is along the tail T of
!                                    |     /            the normal vector.
!                                    |    /
!                                    | A /
!                                    |  /
!                                    | /
!                                    |/
!                                    T (n)
!
!        In case the tilting axis is parallel to the normal vector, the tilting angle becomes
!        undefined and another tilting axis must be chosen. The code below checks this and
!        informs the user.
!
!
     sideTiltingAngle = pi_detectors (detector) % sideTiltingAngle
     sideTiltingAxis  = pi_detectors (detector) % sideTiltingAxis

     sinPhi = sin (sideTiltingAngle)          ! tilting angle is in radians here
     cosPhi = cos (sideTiltingAngle)

     select case (sideTiltingAxis)

     case ('x')

       nxny  = nx * ny
       nxnz  = nx * nz
       nProj = sqrt (ny * ny + nz * nz)

       if (nProj == 0.0) then
           call Driver_abortFlash ("pi_detectorsInfo3DRec: Impossible detector tilting x-axis!")
       else if (nProj < pi_badTiltingAxis) then
           call Logfile_stamp ('Bad detector tilting axis, but will proceed...','[pi_detectorsInfo3DRec]')
       end if

       nProjInv = 1.0 / nProj

       uXx =  - nProj * sinPhi
       uXy = nProjInv * (  nxny * sinPhi  +  nz * cosPhi)
       uXz = nProjInv * (  nxnz * sinPhi  -  ny * cosPhi)

       uYx =    nProj * cosPhi
       uYy = nProjInv * (- nxny * cosPhi  +  nz * sinPhi)
       uYz = nProjInv * (- nxnz * cosPhi  -  ny * sinPhi)

     case ('y')

       nxny  = nx * ny
       nynz  = ny * nz
       nProj = sqrt (nx * nx + nz * nz)

       if (nProj == 0.0) then
           call Driver_abortFlash ("pi_detectorsInfo3DRec: Impossible detector tilting y-axis!")
       else if (nProj < pi_badTiltingAxis) then
           call Logfile_stamp ('Bad detector tilting axis, but will proceed...','[pi_detectorsInfo3DRec]')
       end if

       nProjInv = 1.0 / nProj

       uXx = nProjInv * (  nxny * sinPhi  -  nz * cosPhi)
       uXy =  - nProj * sinPhi
       uXz = nProjInv * (  nynz * sinPhi  +  nx * cosPhi)

       uYx = nProjInv * (- nxny * cosPhi  -  nz * sinPhi)
       uYy =    nProj * cosPhi
       uYz = nProjInv * (- nynz * cosPhi  +  nx * sinPhi)

     case ('z')

       nxnz  = nx * nz
       nynz  = ny * nz
       nProj = sqrt (nx * nx + ny * ny)

       if (nProj == 0.0) then
           call Driver_abortFlash ("pi_detectorsInfo3DRec: Impossible detector tilting z-axis!")
       else if (nProj < pi_badTiltingAxis) then
           call Logfile_stamp ('Bad detector tilting axis, but will proceed...','[pi_detectorsInfo3DRec]')
       end if

       nProjInv = 1.0 / nProj

       uXx = nProjInv * (  nxnz * sinPhi  +  ny * cosPhi)
       uXy = nProjInv * (  nynz * sinPhi  -  nx * cosPhi)
       uXz =  - nProj * sinPhi

       uYx = nProjInv * (- nxnz * cosPhi  +  ny * sinPhi)
       uYy = nProjInv * (- nynz * cosPhi  -  nx * sinPhi)
       uYz =    nProj * cosPhi

     case default

       call Driver_abortFlash ("pi_detectorsInfo3DRec: Invalid detector tilting axis!")

     end select
!
!
!     ...unit vector components are ready. Check for unit vector conditions and if ok
!        store into detectors array.
!
!
     normX   = uXx * uXx  +  uXy * uXy  +  uXz * uXz
     normY   = uYx * uYx  +  uYy * uYy  +  uYz * uYz
     orthoXY = uXx * uYx  +  uXy * uYy  +  uXz * uYz
     orthoXn = uXx *  nx  +  uXy *  ny  +  uXz *  nz
     orthoYn = uYx *  nx  +  uYy *  ny  +  uYz *  nz

     if (     (normX > 1.0 + pi_normalizedTolerance) &
         .or. (normX < 1.0 - pi_normalizedTolerance) ) then
          call Driver_abortFlash ("pi_detectorsInfo3DRec: Axis X unit vector not normalized!")
     end if

     if (     (normY > 1.0 + pi_normalizedTolerance) &
         .or. (normY < 1.0 - pi_normalizedTolerance) ) then
          call Driver_abortFlash ("pi_detectorsInfo3DRec: Axis Y unit vector not normalized!")
     end if

     if (     (orthoXY >   pi_orthogonalTolerance) &
         .or. (orthoXY < - pi_orthogonalTolerance) ) then
          call Driver_abortFlash ("pi_detectorsInfo3DRec: Axis X,Y unit vectors not orthogonal!")
     end if

     if (     (orthoXn >   pi_orthogonalTolerance) &
         .or. (orthoXn < - pi_orthogonalTolerance) ) then
          call Driver_abortFlash ("pi_detectorsInfo3DRec: Axis X + detector normal vectors not orthogonal!")
     end if

     if (     (orthoYn >   pi_orthogonalTolerance) &
         .or. (orthoYn < - pi_orthogonalTolerance) ) then
          call Driver_abortFlash ("pi_detectorsInfo3DRec: Axis Y + detector normal vectors not orthogonal!")
     end if

     pi_detectors (detector) % axisXunitX = uXx
     pi_detectors (detector) % axisXunitY = uXy
     pi_detectors (detector) % axisXunitZ = uXz
     pi_detectors (detector) % axisYunitX = uYx
     pi_detectors (detector) % axisYunitY = uYy
     pi_detectors (detector) % axisYunitZ = uYz

  end do
!
!
!     ...Ready!
!
!
  return
end subroutine pi_detectorsInfo3DRec
