!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonDetection/pi_recordProtonOnScreen
!!
!! NAME
!!
!!  pi_recordProtonOnScreen
!!
!! SYNOPSIS
!!
!!  call pi_recordProtonOnScreen (real    (in)    :: px,
!!                                real    (in)    :: py,
!!                                real    (in)    :: pz,
!!                                real    (in)    :: vx,
!!                                real    (in)    :: vy,
!!                                real    (in)    :: vz,
!!                                integer (in)    :: detector)
!!
!! DESCRIPTION
!!
!!  Records the proton on the specified detector screen. The proton has an initial
!!  position and direction (velocity). The routine records the (x,y) location on
!!  the screen in terms of local screen coordinates. The (x,y) screen coordinates
!!  are rescaled to the square screen side length, such that all protons within
!!  the screen have screen coordinate paris (x,y) within the range [0,1]. Protons
!!  falling outside the screen are also saved for eventual further diagnostic studies.
!!  Protons which miss the screen plane entirely are not recorded.
!!
!!  Procedure:
!!
!!                                    |\
!!                                    | \
!!                                    |  \                   C = detector center
!!                                    |   \
!!                                    | uY|\                 n = normal unit vector
!!                                    |   | \
!!                                    |   |  \         uX,uY,n = form a right hand rule
!!                      detector      |   |   |                  orthogonal vector system:
!!                                    |   C---|---> n                  n = uY x uX
!!                       screen       \  / \  |
!!                                     \/   \ |              P = position of proton
!!                                     /\  uX\|
!!                                    /  \    |              v = velocity vector of proton
!!                                   /    \   |
!!                                  /      \  |          (x,y) = local screen coordinates
!!                                 /        \ |                  (basis = uX,uY) where proton
!!                                /          \|                   will hit the screen plane
!!                               /
!!                              /
!!                             /
!!                            P---> v -----------(x,y)
!!
!!
!!  From geometry:
!!
!!                    i) [(C - P) dot n] / [v dot n] > 0  ->  proton hits screen plane
!!
!!                   ii) x = uY dot [v cross (C - P)] / [v dot n]
!!
!!                  iii) y = uX dot [(C - P) cross v] / [v dot n]
!!
!!                   iv) if [v dot n] = 0  ->  parallel to screen plane, no hit
!!
!!
!! ARGUMENTS
!!
!!  px       : position global x-coordinate of the proton
!!  py       : position global y-coordinate of the proton
!!  pz       : position global z-coordinate of the proton
!!  vx       : velocity x-coordinate component of the proton
!!  vy       : velocity y-coordinate component of the proton
!!  vz       : velocity z-coordinate component of the proton
!!  detector : detector screen number on which to record the proton
!!
!! NOTES
!!
!!***

subroutine pi_recordProtonOnScreen (px, py, pz, vx, vy, vz, detector)

  use ProtonImaging_data,      ONLY : pi_detectors,              &
                                      pi_recordOffscreenProtons, &
                                      pi_screenProtonCount,      &
                                      pi_screenProtons

  use Driver_interface,        ONLY : Driver_abortFlash

  implicit none

#include "constants.h"
#include "Flash.h"
#include "ProtonImaging.h"

  real,    intent (in) :: px, py, pz
  real,    intent (in) :: vx, vy, vz
  integer, intent (in) :: detector

  logical :: onScreen
  logical :: recordProton

  integer :: i,j,k
  integer :: ip,jp,kp
  integer :: proton
  integer :: protonTag

  real    :: cpdotn
  real    :: cpvx, cpvy, cpvz
  real    :: cpx, cpy, cpz
  real    :: nx, ny, nz
  real    :: sHalf, sInv
  real    :: uXx, uXy, uXz
  real    :: uYx, uYy, uYz
  real    :: vdotn, vdotnInv
  real    :: x, y
!
!
!     ...Load the detector specifics.
!
!
  nx  = pi_detectors (detector) % normalX
  ny  = pi_detectors (detector) % normalY
  nz  = pi_detectors (detector) % normalZ

  cpx = pi_detectors (detector) % centerX  -  px
  cpy = pi_detectors (detector) % centerY  -  py
  cpz = pi_detectors (detector) % centerZ  -  pz

  uXx = pi_detectors (detector) % axisXunitX
  uXy = pi_detectors (detector) % axisXunitY
  uXz = pi_detectors (detector) % axisXunitZ

  uYx = pi_detectors (detector) % axisYunitX
  uYy = pi_detectors (detector) % axisYunitY
  uYz = pi_detectors (detector) % axisYunitZ

  sHalf = pi_detectors (detector) % sideLengthHalf
  sInv  = pi_detectors (detector) % sideLengthInv
!
!
!     ...Form the scalar product of v with n and check, if the proton travels parallel to
!        the screen (a miss). If not, calculate its inverse.
!
!
  vdotn = vx * nx + vy * ny + vz * nz

  if (vdotn == 0.0) return

  vdotnInv = 1.0 / vdotn
!
!
!     ...Form the scalar product of (C-P) with n and check, if the proton travels away from
!        the screen (a miss).
!
!
  cpdotn = cpx * nx + cpy * ny + cpz * nz

  if (cpdotn * vdotnInv <= 0.0) return
!
!
!     ...We are now sure the proton hits the screen plane. Form the cross product of (c-p)
!        with v and calculate the (x,y) screen coordinates from the center of the screen.
!
!
  cpvx = cpy * vz - cpz * vy
  cpvy = cpz * vx - cpx * vz
  cpvz = cpx * vy - cpy * vx

  x = - vdotnInv * (uYx * cpvx + uYy * cpvy + uYz * cpvz)
  y =   vdotnInv * (uXx * cpvx + uXy * cpvy + uXz * cpvz)
!
!
!     ...Shift to positive (x,y) pairs within the screen and rescale to [0-1]. Check, if
!        we want this proton recorded. Record also the detector number.
!
!
  x = (x + sHalf) * sInv
  y = (y + sHalf) * sInv

  onScreen = (x >= 0.0 .and. x <= 1.0 .and. y >= 0.0 .and. y <= 1.0)
  recordProton = onScreen .or. pi_recordOffscreenProtons

  if (recordProton) then
      pi_screenProtonCount = pi_screenProtonCount + 1
      pi_screenProtons (SCREEN_POSX,pi_screenProtonCount) = x
      pi_screenProtons (SCREEN_POSY,pi_screenProtonCount) = y
      pi_screenProtons (SCREEN_DETC,pi_screenProtonCount) = real (detector)
  end if
!
!
!     ...Ready!
!
!
  return
end subroutine pi_recordProtonOnScreen
