!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonUtilities/pi_cubeOhIndex2FullSetXYZ
!!
!! NAME
!!
!!  pi_cubeOhIndex2FullSetXYZ
!!
!! SYNOPSIS
!!
!!  call pi_cubeOhIndex2FullSetXYZ (real,    intent (in)  :: s,
!!                                  integer, intent (in)  :: iOh,
!!                                  integer, intent (in)  :: jOh,
!!                                  integer, intent (in)  :: kOh,
!!                                  integer, intent (out) :: nxyz,
!!                                  integer, intent (out) :: xSet (1:48),
!!                                  integer, intent (out) :: ySet (1:48),
!!                                  integer, intent (out) :: zSet (1:48))
!!
!! DESCRIPTION
!!
!!  Given an Oh group index triple (i,j,k) of a cube with side length s as input, this routine
!!  returns the full set of local x,y,z coordinates associated with that index that can be
!!  obtained by applying Oh group symmetry operations.
!!
!!  Problem setup:
!!  -------------
!!
!!  The following setup of the problem is assumed:
!!
!!      1) A 3D cartesian coordinate system is placed in space.
!!      2) The following index -> coordinate association is made:
!!
!!                     i -> index on local x-axis
!!                     j -> index on local y-axis
!!                     k -> index on local z-axis
!!
!!      3) Cubes of side length s are placed along the coordinate system with one vertex of
!!         each of the 8 innermost cubes located at the origin (0,0,0) of the cartesian
!!         coordinate system.
!!      4) The tics (the i,j,k) on the coordinate axes correspond to the cubes vertices.
!!      5) Each cube can be labeled by an index triple (i,j,k), where (i,j,k) denotes the
!!         absoulte largest number of all the 8 vertex indices.
!!      6) The location inside each cube where the x,y,z coordinates are going to be
!!         determined is the center of each cube.
!!
!!  Symmetry considerations:
!!  -----------------------
!!
!!  Since the cartesian coordinate system has Oh-group symmetry, each (i,j,k) index has
!!  potentially up 48 symmetry related indices (including itself):
!!
!!                (i,j,k) -> 6 permutations x 8 sign maps = 48
!!
!!  The six permutations are:   (i,j,k)    The eight sign maps are:  + + +
!!                              (i,k,j)                              + + -
!!                              (j,i,k)                              + - +
!!                              (j,k,i)                              - + +
!!                              (k,i,j)                              + - -
!!                              (k,j,i)                              - + -
!!                                                                   - - +
!!                                                                   - - -
!!
!!  The number of permutations are smaller, if index equalities are present, but the
!!  eight sign maps remain the same. Note, that from the definition of the (i,j,k) cube
!!  index (outermost vertex), indices containing 0's are not allowed.
!!
!! ARGUMENTS
!!
!!  s     : the cube's side length
!!  iOh   : the i-part of the cube's Oh index (i,j,k)
!!  jOh   : the j-part of the cube's Oh index (i,j,k)
!!  kOh   : the k-part of the cube's Oh index (i,j,k)
!!  nxyz  : the number of total (x,y,z) coordinates returned (up to 48)
!!  xSet  : contains the full set of all the x-coordinates
!!  ySet  : contains the full set of all the y-coordinates
!!  zSet  : contains the full set of all the z-coordinates
!!
!! NOTES
!!
!!  The routine checks, if the indices transmitted are within the expected range.
!!
!!***

subroutine pi_cubeOhIndex2FullSetXYZ (s,                   &
                                      iOh, jOh, kOh,       &
                                                     nxyz, &
                                                     xSet, &
                                                     ySet, &
                                                     zSet  )

  use Driver_interface,  ONLY : Driver_abortFlash

  implicit none

  real,    intent (in)    :: s
  integer, intent (in)    :: iOh, jOh, kOh
  integer, intent (out)   :: nxyz
  real,    intent (out)   :: xSet (1:48)
  real,    intent (out)   :: ySet (1:48)
  real,    intent (out)   :: zSet (1:48)

  integer, save :: istart, jstart, kstart

  integer :: caseOhIndex
  integer :: i,j,k
  integer :: m,m2,m3,m4,m5,m6,m7,m8

  real    :: x,y,z
!
!
!     ...Check, if indices are in proper range.
!
!
  if (iOh == 0 .or. jOh == 0 .or. kOh == 0) then
      call Driver_abortFlash ("pi_cubeOhIndex2FullSetXYZ: A cube Oh index with zeros is illegal!")
  end if
!
!
!     ...Analyze the Oh index and proceed accordingly. After this section we have
!        three possible cases:
!
!              caseOhIndex = 1 -> all indices i,j,k mutually different
!              caseOhIndex = 2 -> only two indices i,j different
!              caseOhIndex = 3 -> all indices equal (i will be used)
!
!
  i = abs (iOh)
  j = abs (jOh)
  k = abs (kOh)

  caseOhIndex = 2

  if (i /= j) then
      if (j /= k .and. i /= k) caseOhIndex = 1 
  else
      if (j /= k) then
          j = k
      else
          caseOhIndex = 3
      end if
  end if
!
!
!     ...Handle all three cases.
!
!
  select case (caseOhIndex)

  case (1)  ! all indices i,j,k mutually different

    m = 6
    x = s * (real (i) - 0.5)
    y = s * (real (j) - 0.5)
    z = s * (real (k) - 0.5)

    xSet (1) = x
    xSet (2) = x
    xSet (3) = y
    xSet (4) = y
    xSet (5) = z
    xSet (6) = z

    ySet (1) = y
    ySet (2) = z
    ySet (3) = x
    ySet (4) = z
    ySet (5) = x
    ySet (6) = y

    zSet (1) = z
    zSet (2) = y
    zSet (3) = z
    zSet (4) = x
    zSet (5) = y
    zSet (6) = x

  case (2)      ! only two indices i,j different

    m = 3
    x = s * (real (i) - 0.5)
    y = s * (real (j) - 0.5)

    xSet (1) = x
    xSet (2) = y
    xSet (3) = y

    ySet (1) = y
    ySet (2) = x
    ySet (3) = y

    zSet (1) = y
    zSet (2) = y
    zSet (3) = x

  case (3)      ! all indices equal

    m = 1
    x = s * (real (i) - 0.5)
    xSet (1) = x
    ySet (1) = x
    zSet (1) = x

  end select
!
!
!     ...Expand to complete set including sign permutations.
!
!
  m2 =  m + m
  m3 = m2 + m
  m4 = m3 + m
  m5 = m4 + m
  m6 = m5 + m
  m7 = m6 + m
  m8 = m7 + m

  xSet ( m+1:m2) =  + xSet (1:m)
  xSet (m2+1:m3) =  + xSet (1:m)
  xSet (m3+1:m4) =  - xSet (1:m)
  xSet (m4+1:m5) =  + xSet (1:m)
  xSet (m5+1:m6) =  - xSet (1:m)
  xSet (m6+1:m7) =  - xSet (1:m)
  xSet (m7+1:m8) =  - xSet (1:m)

  ySet ( m+1:m2) =  + ySet (1:m)
  ySet (m2+1:m3) =  - ySet (1:m)
  ySet (m3+1:m4) =  + ySet (1:m)
  ySet (m4+1:m5) =  - ySet (1:m)
  ySet (m5+1:m6) =  + ySet (1:m)
  ySet (m6+1:m7) =  - ySet (1:m)
  ySet (m7+1:m8) =  - ySet (1:m)

  zSet ( m+1:m2) =  - zSet (1:m)
  zSet (m2+1:m3) =  + zSet (1:m)
  zSet (m3+1:m4) =  + zSet (1:m)
  zSet (m4+1:m5) =  - zSet (1:m)
  zSet (m5+1:m6) =  - zSet (1:m)
  zSet (m6+1:m7) =  + zSet (1:m)
  zSet (m7+1:m8) =  - zSet (1:m)

  nxyz = m8
!
!
!     ...Ready!
!
!
  return
end subroutine pi_cubeOhIndex2FullSetXYZ
