!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonUtilities/pi_cubesInSphereOhIndices
!!
!! NAME
!!
!!  pi_cubesInSphereOhIndices
!!
!! SYNOPSIS
!!
!!  call pi_cubesInSphereOhIndices (integer, intent (in)    :: n,
!!                                  integer, intent (in)    :: maxIndices,
!!                                  logical, intent (inout) :: startIndices,
!!                                  logical, intent (inout) :: moreIndices,
!!                                  integer, intent (out)   :: nIndices,
!!                                  integer, intent (out)   :: indexI (1:maxIndices),
!!                                  integer, intent (out)   :: indexJ (1:maxIndices),
!!                                  integer, intent (out)   :: indexK (1:maxIndices))
!!
!! DESCRIPTION
!!
!!  This routine retrieves a set of Oh group (symmetry) indices of cubes located inside a sphere.
!!
!!  Problem setup:
!!  -------------
!!
!!  The following setup of the problem is assumed:
!!
!!      1) A 3D cartesian coordinate system is placed in space.
!!      1) A sphere is placed at the cartesian origin (0,0,0), having exactly n equal sized
!!         divisions along its radius R in all three cartesian directions.
!!      2) Cubes of side size R/n are placed inside the sphere with one vertex of each of
!!         the 8 innermost cubes located at the origin (0,0,0) of the sphere and the cartesian
!!         coordinate system.
!!      3) Each cube can be labeled by three indices (i,j,k), where (i,j,k) denotes the
!!         largest number of the vertex indices. A vertex index simply denotes the division
!!         number on which it is placed in the coordiate system.
!!
!!  As an example, consider the innermost cube its 8 vertex indices (0,0,0), (0,0,1), (0,1,0),
!!  (1,0,0), (0,1,1), (1,0,1), (1,1,0) and (1,1,1). The cube's labelling index would then be
!!  the last vertex index (1,1,1).
!!
!!  Each cube's labelling indices is therefore an indication of how far out it reaches in space.
!!  To determine if a cube is located inside the sphere, its labelling indices (i,j,k) must
!!  obey the following inequality:
!!
!!                        (iR/n)^2 + (jR/n)^2 + (kR/n)^2 <= R^2
!!
!!  which (after rescaling to unit radius of the sphere) is the same as:
!!
!!                              i^2 + j^2 + k^2 <= n^2       (I)
!!
!!  Symmetry considerations:
!!  -----------------------
!!
!!  Since the cartesian coordinate system (and the sphere) have Oh-group symmetry, only a subset
!!  of the labelling indices needs to be found. The others can be determined from symmetry
!!  operations in the Oh-group. The following set of labelling indices (Oh indices) need
!!  to be evaluated, which obey the above Diophantine equation (I):
!!
!!                       i,j,k >= 1   (only positive indices, one octant only)
!!                      k >= j >= i   (unique tetrahedron in octant)
!!
!!  Performance considerations:
!!  --------------------------
!!
!!  Since the number of such Oh indices may be very large even for moderately sized n, the
!!  collection of unique indices is returned in batches of user defined size. Several consecutive
!!  calls to the routine thus retrieves the complete set of Oh indices.
!!
!! ARGUMENTS
!!
!!  n            : number of divisions along the sphere's radius
!!  maxIndices   : the maximum number of Oh indices that can be returned per call
!!  startIndices : if true, the index retrieval starts from the beginning 
!!  moreIndices  : if true, more Oh indices are expected
!!  nIndices     : the actual number of Oh indices returned
!!  index(I,J,K) : the i,j,k-parts of the Oh index (i,j,k)
!!
!! NOTES
!!
!!  This routine does not need to know in advance the total number of Oh indices.
!!  It just retrieves the Oh indices until no more are found. Any check for correctness
!!  in the number of overall retrieved Oh indices is left to be done outside of this
!!  routine.
!!
!!***

subroutine pi_cubesInSphereOhIndices (n,                          &
                                      maxIndices,                 &
                                      startIndices,               &
                                                     moreIndices, &
                                                     nIndices,    &
                                                     indexI,      &
                                                     indexJ,      &
                                                     indexK       )

  implicit none

  integer, intent (in)    :: n
  integer, intent (in)    :: maxIndices
  logical, intent (inout) :: startIndices
  logical, intent (inout) :: moreIndices
  integer, intent (out)   :: nIndices
  integer, intent (out)   :: indexI (1:maxIndices)
  integer, intent (out)   :: indexJ (1:maxIndices)
  integer, intent (out)   :: indexK (1:maxIndices)

  integer, save :: istart, jstart, kstart

  integer :: i,j,k,m,p,q,r
!
!
!     ...If requested, start the retrieval of the Oh indices from the beginning.
!
!
  if (startIndices) then
      istart = 1
      jstart = 1
      kstart = 1
      startIndices = .false.
  end if
!
!
!     ...Retrieve the current array of Oh indices.
!
!
  m = 0
  p = n * n - (kstart - 1) * (kstart - 1)

  do k = kstart,n-1
     p = p - k - k + 1                            ! accumulate n^2 - k^2
     q = p - (jstart - 1) * (jstart - 1)
     do j = jstart,k
        q = q - j - j + 1                         ! accumulate n^2 - k^2 - j^2
        if (q > 0) then
            r = q - (istart - 1) * (istart - 1)
            do i = istart,j
               r = r - i - i + 1                  ! accumulate n^2 - k^2 - j^2 - i^2

               if (r > 0) then
                   m = m + 1
                   if (m <= maxIndices) then
                       indexI (m) = i
                       indexJ (m) = j
                       indexK (m) = k
                   else
                       nIndices = maxIndices      ! current Oh index batch is full
                       istart = i                 ! save latest i,j,k for next
                       jstart = j                 ! call of routine
                       kstart = k
                       return
                   end if
               else
                   exit
               end if

            end do
            istart = 1
        else
            exit
        end if
     end do
     jstart = 1
  end do

  moreIndices = .false.
!
!
!     ...Ready!
!
!
  return
end subroutine pi_cubesInSphereOhIndices
