!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonUtilities/pi_printBeamsData
!!
!! NAME
!!
!!  pi_printBeamsData
!!
!! SYNOPSIS
!!
!!  call pi_printBeamsData ()
!!
!! DESCRIPTION
!!
!!  Utility routine, which prints detailed info regarding the generated data
!!  for all beams to a text file. The information is written out to a file named
!!  <basenm>ProtonBeamsPrint.txt, where <basenm> is the runtime parameter for
!!  output file names.
!!
!!***

subroutine pi_printBeamsData ()

  use Driver_interface,    ONLY : Driver_abortFlash

  use ProtonImaging_data,  ONLY : pi_baseName,      &
                                  pi_beams,         &
                                  pi_globalMe,      &
                                  pi_numberOfBeams, &
                                  pi_speedOfLight

  implicit none
   
#include "constants.h"
#include "ProtonImaging.h"
#include "Flash.h"

  character (len =  MAX_STRING_LENGTH) :: fileName

  logical :: ignoreBnd

  integer :: beam
  integer :: dimensions
  integer :: fileUnit
  integer :: nProtons
  integer :: seed, seedMaximum, seedStepping
  integer :: ut_getFreeFileUnit

  real    :: c
  real    :: apertureAngle
  real    :: axisUnit1X, axisUnit1Y, axisUnit1Z
  real    :: axisUnit2X, axisUnit2Y, axisUnit2Z
  real    :: distL2T
  real    :: lensRadius
  real    :: lensX, lensY, lensZ
  real    :: magnifyT2L
  real    :: protonEnergy, protonSpeed
  real    :: targetRadius
  real    :: targetX, targetY, targetZ
!
!
!   ...Do the printout only on the master processor.
!
!
  if (pi_globalMe /= MASTER_PE) then
      return
  end if
!
!
!   ...Open the printout file.
!
!
  fileUnit = ut_getFreeFileUnit ()
  fileName = trim (pi_baseName) // "ProtonBeamsPrint.txt"

  open (fileUnit, file = fileName)
!
!
!     ...Loop over all beams.
!
!
  c = pi_speedOfLight

  do beam = 1, pi_numberOfBeams

     apertureAngle = pi_beams (beam) % apertureAngle
     axisUnit1X    = pi_beams (beam) % axisUnit1X
     axisUnit1Y    = pi_beams (beam) % axisUnit1Y
     axisUnit1Z    = pi_beams (beam) % axisUnit1Z
     axisUnit2X    = pi_beams (beam) % axisUnit2X
     axisUnit2Y    = pi_beams (beam) % axisUnit2Y
     axisUnit2Z    = pi_beams (beam) % axisUnit2Z
     dimensions    = pi_beams (beam) % dimensionality
     distL2T       = pi_beams (beam) % distanceLens2Target
     seed          = pi_beams (beam) % gridSeed
     seedMaximum   = pi_beams (beam) % gridSeedMaximum
     seedStepping  = pi_beams (beam) % gridSeedStepping
     ignoreBnd     = pi_beams (beam) % noBoundaryCondition
     protonSpeed   = pi_beams (beam) % initialProtonSpeed
     lensRadius    = pi_beams (beam) % lensRadius
     lensX         = pi_beams (beam) % lensX
     lensY         = pi_beams (beam) % lensY
     lensZ         = pi_beams (beam) % lensZ
     nProtons      = pi_beams (beam) % numberOfProtons
     protonEnergy  = pi_beams (beam) % protonEnergy
     magnifyT2L    = pi_beams (beam) % target2LensMagnification
     targetRadius  = pi_beams (beam) % targetRadius
     targetX       = pi_beams (beam) % targetX
     targetY       = pi_beams (beam) % targetY
     targetZ       = pi_beams (beam) % targetZ

     write (fileUnit,'(/)'        )
     write (fileUnit,'(a,i2)'     ) "               PROTON BEAM NR ",beam
     write (fileUnit,'(/)'        )
     write (fileUnit,'(a,es20.12)') " Beam aperture angle (rad)                      = ", apertureAngle
     write (fileUnit,'(a,es20.12)') " Distance lens center --> target center         = ", distL2T
     write (fileUnit,'(a,es20.12)') " Proton energy (in MeV)                         = ", protonEnergy
     write (fileUnit,'(a,es20.12)') " Initial proton speed (in units of cm/s)        = ", protonSpeed
     write (fileUnit,'(a,es20.12)') " Initial proton speed (in units of light speed) = ", protonSpeed / c
     write (fileUnit,'(a,i20)'    ) " Number of protons in beam                      = ", nProtons
     write (fileUnit,'(a,L1)'     ) " The domain boundary conditions are ignored     = ", ignoreBnd

     select case (dimensions)

     case (3)

     write (fileUnit,'(a,i20)'    ) " Current seed value for random grid             = ", seed
     write (fileUnit,'(a,i20)'    ) " Maximum seed value for random grid             = ", seedMaximum
     write (fileUnit,'(a,i20)'    ) " Stepping seed value for random grid            = ", seedStepping
     write (fileUnit,'(a,es20.12)') " Lens radius length                             = ", lensRadius
     write (fileUnit,'(a,es20.12)') " Lens global x-coordinate                       = ", lensX
     write (fileUnit,'(a,es20.12)') " Lens global y-coordinate                       = ", lensY
     write (fileUnit,'(a,es20.12)') " Lens global z-coordinate                       = ", lensZ
     write (fileUnit,'(a,es20.12)') " 1st axis unit vector x-coordinate              = ", axisUnit1X
     write (fileUnit,'(a,es20.12)') " 1st axis unit vector y-coordinate              = ", axisUnit1Y
     write (fileUnit,'(a,es20.12)') " 1st axis unit vector z-coordinate              = ", axisUnit1Z
     write (fileUnit,'(a,es20.12)') " 2nd axis unit vector x-coordinate              = ", axisUnit2X
     write (fileUnit,'(a,es20.12)') " 2nd axis unit vector y-coordinate              = ", axisUnit2Y
     write (fileUnit,'(a,es20.12)') " 2nd axis unit vector z-coordinate              = ", axisUnit2Z
     write (fileUnit,'(a,es20.12)') " Target radius length                           = ", targetRadius
     write (fileUnit,'(a,es20.12)') " Target global x-coordinate                     = ", targetX
     write (fileUnit,'(a,es20.12)') " Target global y-coordinate                     = ", targetY
     write (fileUnit,'(a,es20.12)') " Target global z-coordinate                     = ", targetZ
     write (fileUnit,'(a,es20.12)') " Target -> lens magnification factor            = ", magnifyT2L
     write (fileUnit,'(a)'        ) " ------------------------------------------------ "

     case default

     call Driver_abortFlash ("pi_printBeamsData: Bad dimensionality of beam (programmers fault)!")

     end select

  end do
!
!
!   ...Close the printout file.
!
!
  close (fileUnit)
!
!
!    ...Ready!
!
!  
  return
end subroutine pi_printBeamsData
