!!****if* source/diagnostics/ProtonImaging/ProtonImagingMain/ProtonsTrace/pi_blockData3DRec
!!
!! NAME
!!
!!  pi_blockData3DRec
!!
!! SYNOPSIS
!!
!!  call pi_blockData3DRec (integer (in) :: iminBlock,
!!                          integer (in) :: imaxBlock,
!!                          integer (in) :: jminBlock,
!!                          integer (in) :: jmaxBlock,
!!                          integer (in) :: kminBlock,
!!                          integer (in) :: kmaxBlock,
!!                          real    (in) :: blockData (:,:,:,:))
!!
!! DESCRIPTION
!!
!!  Computes cell data for one specific block for those geometries consisting formally of
!!  3D rectangular grids (cartesian). The block is specified through its number ID and the
!!  blockData array, which contains the needed data for the block. The following is computed
!!  and stored into specific arrays:
!!
!!     1) the cell center electric field components (Ex,Ey,Ex)
!!     2) the cell center magnetic flux density components (Bx,By,Bx)
!!     3) the cell boundary indicator
!!
!!  The necessary arrays for storage must have been allocated before calling this routine.
!!  No checks are done on the passed cell index limits. This is done before calling this
!!  routine.
!!
!! ARGUMENTS
!!
!!  iminBlock : minimum cell i-index limit defining the interior block
!!  imaxBlock : maximum cell i-index limit defining the interior block
!!  jminBlock : minimum cell j-index limit defining the interior block
!!  jmaxBlock : maximum cell j-index limit defining the interior block
!!  kminBlock : minimum cell k-index limit defining the interior block
!!  kmaxBlock : maximum cell k-index limit defining the interior block
!!  blockData : four-dimensional array containing the block data
!!
!!***

subroutine pi_blockData3DRec (iminBlock, imaxBlock, &
                              jminBlock, jmaxBlock, &
                              kminBlock, kmaxBlock, &
                              blockData             )

  use Driver_interface,    ONLY : Driver_abortFlash

  use ProtonImaging_data,  ONLY : pi_cellBfield,            &
                                  pi_cellBoundary,          &
                                  pi_cellEfield,            &
                                  pi_ignoreElectricalField, &
                                  pi_speedOfLightInv,       &
                                  pi_squareRoot4Pi

  implicit none

#include "constants.h"
#include "Flash.h"

  integer, intent (in) :: iminBlock, imaxBlock
  integer, intent (in) :: jminBlock, jmaxBlock
  integer, intent (in) :: kminBlock, kmaxBlock
  real,    intent (in) :: blockData (:,:,:,:)

  integer :: i,j,k

  real    :: factor
!
!
!     ...Get all needed cell quantities.
!
!
  if (pi_ignoreElectricalField) then

      do k = kminBlock,kmaxBlock
         do j = jminBlock,jmaxBlock
            do i = iminBlock,imaxBlock

               pi_cellEfield (1,i,j,k) = 0.0
               pi_cellEfield (2,i,j,k) = 0.0
               pi_cellEfield (3,i,j,k) = 0.0
               pi_cellBfield (1,i,j,k) = blockData (MAGX_VAR,i,j,k)
               pi_cellBfield (2,i,j,k) = blockData (MAGY_VAR,i,j,k)
               pi_cellBfield (3,i,j,k) = blockData (MAGZ_VAR,i,j,k)
               pi_cellBoundary (i,j,k) = blockData (BDRY_VAR,i,j,k)

            enddo
         enddo
      enddo

  else

      do k = kminBlock,kmaxBlock
         do j = jminBlock,jmaxBlock
            do i = iminBlock,imaxBlock

               pi_cellEfield (1,i,j,k) = blockData (ELEX_VAR,i,j,k)
               pi_cellEfield (2,i,j,k) = blockData (ELEY_VAR,i,j,k)
               pi_cellEfield (3,i,j,k) = blockData (ELEZ_VAR,i,j,k)
               pi_cellBfield (1,i,j,k) = blockData (MAGX_VAR,i,j,k)
               pi_cellBfield (2,i,j,k) = blockData (MAGY_VAR,i,j,k)
               pi_cellBfield (3,i,j,k) = blockData (MAGZ_VAR,i,j,k)
               pi_cellBoundary (i,j,k) = blockData (BDRY_VAR,i,j,k)

            enddo
         enddo
      enddo

  end if
!
!
!     ...Prepare the magnetic flux densities to be used with the CGS form
!        of the Lorentz equation: F = qE + q(v/c) X B. We move the extra
!        1/c factor to the B and also convert the B in FLASH to Gaussian
!        units:
!
!                   B -> B * (1/c) * sqrt (4pi)
!
!
  factor = pi_speedOfLightInv * pi_squareRoot4Pi

  pi_cellBfield (:,:,:,:) = factor * pi_cellBfield (:,:,:,:)
!
!
!     ...Ready!
!
!
  return
end subroutine pi_blockData3DRec
