!!****if* source/numericalTools/RungeKutta/RungeKuttaMain/save/rk_derivativesODEvector
!!
!! NAME
!!
!!  rk_derivativesODEvector
!!
!! SYNOPSIS
!!
!!  call rk_derivativesODEvector (function, intent (in)  :: f,
!!                                integer,  intent (in)  :: nx,
!!                                integer,  intent (in)  :: m,
!!                                real,     intent (in)  :: x,
!!                                real,     intent (in)  :: y      (:),
!!                                real,     intent (out) :: dmydxm (:))
!!
!! DESCRIPTION
!!
!!  This routine calculates numerically the m-th derivatives of the supplied dependent
!!  variables vector at the value 'x' of the independent variable. If m = 1, then this
!!  is equal to the ODE function 'f', since 'f' caluclates per definition all 1st order
!!  derivatives.
!!
!!  Second order central difference formulas are used for evaluating the derivatives.
!!  In order to minimize loss of accuracy due to adding opposite signed terms, all
!!  positive and negative terms are first summed up individually and only one global
!!  subtraction is done at the end.
!!
!! ARGUMENTS
!!
!!  f      : the ODE function, which defines the ODE system of 1st derivatives
!!  nx     : the number of expected dependent variables
!!  m      : the derivative order requested
!!  x      : the independent variable value at which the n-th derivatives are wanted
!!  y      : the values of the dependent variable(s)
!!  dmydxm : the m-th derivatives at 'x' of all dependent variable(s)
!!
!! NOTES
!!
!!  Contains an interfaced function in the declarations (see below).
!!
!!***

subroutine rk_derivativesODEvector (f, nx, m,    &
                                    x, y,        &
                                          dmydxm )

  use RungeKutta_data,   ONLY : rk_cubeRootMacheps

  use Driver_interface,  ONLY : Driver_abortFlash

  implicit none

  interface
    function f (x,y)
      real, intent (in) :: x
      real, intent (in) :: y (:)
      real              :: f (1:size (y))
    end function f
  end interface

  integer, intent (in)  :: nx
  integer, intent (in)  :: m
  real,    intent (in)  :: x
  real,    intent (in)  :: y      (:)
  real,    intent (out) :: dmydxm (:)

  integer :: d,i,k,n

  real    :: c,h,p,q

  real    :: addVec (1:nx)
  real    :: negSum (1:nx)
  real    :: posSum (1:nx)
  real    :: zero   (1:nx)

  real    :: v (1:nx,1:m)
!
!
!   ...Check, if the routine can handle the request.
!
!
  if (m < 1) then
      call Driver_abortFlash ('[rk_derivativesODEvector] ERROR: derivative order < 1!')
  end if

  n = size (y)

  if (n > nx .or. n > size (dmydxm)) then
      call Driver_abortFlash ('[rk_derivativesODEvector] ERROR: too many dependent variables!')
  end if
!
!
!   ...Handle special situation.
!
!
  if (m == 1) then
      dmydxm (1:n) = f (x, y (1:n))
      return
  end if
!
!
!   ...Higher than 1st order derivatives are requested.
!
!
  if (1 == 0) then

  h = max (abs (x), 1.0) * rk_cubeRootMacheps    ! estimate optimum numerical derivative step
  h = 1.0e-8

  do i = 1,m
     p = (i + i - m - 1) * h
     v (1:n,i) = f (x + p, y (1:n) + p * f (x, y (1:n)))
  end do

  do k = 1,m-2
  do i = 1,m-k
     v (1:n,i) = (v (1:n,i+1) - v (1:n,i)) / (h + h)
  end do
  end do

  dmydxm (1:n) = (v (1:n,2) - v (1:n,1)) / (h + h)

  end if
!
!
!   ...Higher than 1st order derivatives are requested.
!
!
!  if (1 == 0) then

  posSum (1:n) = 0.0
  negSum (1:n) = 0.0
  zero   (1:n) = 0.0

  h = max (abs (x), 1.0) * rk_cubeRootMacheps    ! estimate optimum numerical derivative step
  h = 1.0e-4
  d = mod (m - 1, 2)
  k = (m - 1 - d) / 2

  write (*,*) ' h = ',h
  write (*,*) ' k = ',k
  write (*,*) ' d = ',d

  if (d == 1) then

      c = 0.5               ! this is coeff for 1st order derivatives at x +/- (k+1)h
      p = (k + 1) * h
 
      addVec (1:n) = c * f (x + p, y (1:n) + p * f (x, y (1:n)))
      posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
      negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

      addVec (1:n) = - c * f (x - p, y (1:n) - p * f (x, y (1:n)))
      posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
      negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

      do i = k,1,-1

         c = - real (i*(k+i+2)) * c / real ((k-i+1)*(i+1))
         p = i * h

         write (*,*) ' c = ',c

         addVec (1:n) = c * f (x + p, y (1:n) + p * f (x, y (1:n)))
         posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
         negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

         addVec (1:n) = - c * f (x - p, y (1:n) - p * f (x, y (1:n)))
         posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
         negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

      end do

  else

      c = 1.0               ! this is coeff for 1st order derivatives at x +/- kh
      p = k * h

      addVec (1:n) = c * f (x + p, y (1:n) + p * f (x, y (1:n)))
      posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
      negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

      addVec (1:n) = c * f (x - p, y (1:n) - p * f (x, y (1:n)))
      posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
      negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

      do i = k-1,1,-1

         c = - real (k+i+1) * c / real (k-i)
         p = i * h

         addVec (1:n) = c * f (x + p, y (1:n) + p * f (x, y (1:n)))
         posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
         negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

         addVec (1:n) = c * f (x - p, y (1:n) - p * f (x, y (1:n)))
         posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
         negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

      end do

      c = - real (k+1) * c / real (k)

      addVec (1:n) = c * f (x, y (1:n))
      posSum (1:n) = posSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) > 0.0)
      negSum (1:n) = negSum (1:n) + merge (addVec (1:n), zero (1:n), addVec (1:n) < 0.0)

  end if

  do i = 1,n
     write (*,*) ' pos neg (',i,') = ',posSum (i) , negSum (i)
  end do

  dmydxm (1:n) = (posSum (1:n) + negSum (1:n)) / (h ** (k+k+d))

!  end if
!
!
!   ...Ready! 
!
!
  return
end subroutine rk_derivativesODEvector
