!!****if* source/numericalTools/RungeKutta/RungeKuttaMain/save/rk_stepEAC
!!
!! NAME
!!
!!  rk_stepEAC
!!
!! SYNOPSIS
!!
!!  call rk_stepEAC (function, intent (in)  :: f,
!!                   integer,  intent (in)  :: n,
!!                   integer,  intent (in)  :: nc,
!!                   integer,  intent (in)  :: p,
!!                   integer,  intent (in)  :: s,
!!                   real,     intent (in)  :: x,
!!                   real,     intent (in)  :: y     (:),
!!                   function, intent (in)  :: ymin,
!!                   function, intent (in)  :: ymax,
!!                   real,     intent (in)  :: eFrac,
!!                   real,     intent (in)  :: eBase (:),
!!                   real,     intent (in)  :: htry,
!!                   real,     intent (out) :: hused,
!!                   real,     intent (out) :: hnext,
!!                   real,     intent (out) :: yout  (:),
!!                   real,     intent (out) :: eout  (:))
!!
!! DESCRIPTION
!!
!!  This routine performs a single (E)mbedded (A)daptive (C)onfined Runge Kuta (RK)
!!  step. The type of embedded adaptive step performed is determined by the previously copied
!!  Butcher tableau a,b,c values. The Butcher tableau copied contains implicitly the order,
!!  its size and the kind of adapted step to be performed.
!!
!!  The details of the ODE function are contained in the passed array function 'f', which
!!  constitutes the link of the general Runge Kutta module to the individual applications.
!!  Thus 'f' has to be provided by an application outside of the Runge Kutta unit.
!!
!!  The code performs a confined RK step. The confinement boundaries are (potentially) dependent
!!  on the current value of the dependent variables at the intermediate and final RK points. This
!!  boundary dependency has to be transmitted via via 2 functions 'ymin' and 'ymax', which accept
!!  the current dependent variables array as argument input. The user has to provide the external
!!  definition of these 2 functions.
!!
!!  The code checks if at any stage during the RK step the intermediate RK points (i.e. dependent
!!  variables) are out of bounds for the confined dependent variables. If the case, the step
!!  size must be reduced (currently by a fixed factor < 1, but this factor might be passed in
!!  the future as argument as well).
!!
!!  Note that the layout of the dependent variable array can always be done such that the
!!  confined variables come first. The user must therefore be careful to design its ODE
!!  function in such a way that those confined dependent variables are at the beginning
!!  of the dependent variable array.
!!
!! ARGUMENTS
!!
!!  f            : the ODE function containing details of the ODE system to be solved
!!  n            : the number of dependent variables
!!  nc           : the number of confined dependent variables (nc <= n, not checked here!)
!!  p            : the order of the lowest order RK step (error scales as O[h^(p+1)])
!!  s            : the Butcher tableau size
!!  x            : the initial value of the independent variable
!!  y            : the initial values of the dependent variable(s)
!!  ymin         : the lower bound confinement function (as function of dependent variables)
!!  ymax         : the upper bound confinement function (as function of dependent variables)
!!  eFrac        : the fractional value for each of the error base values
!!  eBase        : the error base values for each of the dependent variables
!!  htry         : the initial step size on the independent variable
!!  hused        : the final step size applied on the independent variable
!!  hnext        : the suggested next step size on the independent variable
!!  yout         : the values of the dependent variable(s) after the accepted RK step
!!  eout         : the returned error of all dependent variable(s)
!!
!! NOTES
!!
!!  Contains an interfaced function in the declarations (see below).
!!
!!***

subroutine rk_stepEAC (f, n, nc, p, s,               &
                       x, y, ymin, ymax,             &
                       eFrac, eBase,                 &
                       htry,                         &
                                       hused, hnext, &
                                       yout, eout    )

  use RungeKutta_data,   ONLY : rk_a,                         &
                                rk_b,                         &
                                rk_c,                         &
!                                rk_kVectors,                  &
                                rk_stepSizeConfinementFactor, &
                                rk_stepSizeSafetyFactor

  use Driver_interface,  ONLY : Driver_abortFlash

  implicit none

  interface
    function f (x,y)
      real, intent (in) :: x
      real, intent (in) :: y (:)
      real              :: f (1:size (y))
    end function f
  end interface

  interface
    function ymin (nc,y)
      integer, intent (in) :: nc
      real,    intent (in) :: y (:)
      real                 :: ymin (1:nc)
    end function ymin
  end interface

  interface
    function ymax (nc,y)
      integer, intent (in) :: nc
      real,    intent (in) :: y (:)
      real                 :: ymax (1:nc)
    end function ymax
  end interface

  integer, intent (in)  :: n
  integer, intent (in)  :: nc
  integer, intent (in)  :: p
  integer, intent (in)  :: s
  real,    intent (in)  :: x
  real,    intent (in)  :: y     (:)
  real,    intent (in)  :: eFrac
  real,    intent (in)  :: eBase (:)
  real,    intent (in)  :: htry
  real,    intent (out) :: hused
  real,    intent (out) :: hnext
  real,    intent (out) :: yout  (:)
  real,    intent (out) :: eout  (:)

  logical :: badStep
  logical :: stagnation

  integer :: i,j

  real    :: eMax
  real    :: eRescale
  real    :: h

  real    :: fsave (1:n)
  real    :: rk_kVectors (1:n,1:s)
!
!
!   ...Do the embedded adaptive confined Runge Kutta step.
!
!
  h = htry

! write (*,*) ' h initial = ',h

  badStep = .true.
  eRescale = - 1.0 / real (p + 1)

  fsave (1:n) = f (x, y (1:n))         ! save initial function (no recompute, if h changes)

  stepSizeLoop: do while (badStep)

     rk_kVectors (1:n,1) = h * fsave (1:n)

     do i = 2,s
        yout (1:n) = y (1:n)                                              ! temporarily use the 'yout'
        do j = 1,i-1                                                      ! vectors for accumulating
           yout (1:n) = yout (1:n) + rk_a (i,j) * rk_kVectors (1:n,j)     ! the k vectors for error
        end do                                                            ! analysis

!  write (*,'(a,es30.16)') ' x    = ',yout(1)
!  write (*,'(a,es30.16)') ' xmin = ',ymin (1, yout (1:n))
!  write (*,'(a,es30.16)') ' xmax = ',ymax (1, yout (1:n))
!  write (*,'(a,es30.16)') ' 1(xmax-xmin) = ',1.0/(ymax (1, yout (1:n))-ymin (1, yout (1:n)))
!  write (*,'(a,es30.16)') ' true value = ',(yout(1)-ymin (1, yout (1:n)))/(ymax (1, yout (1:n))-ymin (1, yout (1:n))) 


        if (any (yout (1:nc) < ymin (nc, yout (1:n)) .or. &               ! check, if intermediate
                 yout (1:nc) > ymax (nc, yout (1:n)))     ) then          ! RK points step outside
            h = h * rk_stepSizeConfinementFactor                          ! the confinement barriers
!  write (*,*) ' out of confinement, new h = ',h
            cycle stepSizeLoop
        end if

        rk_kVectors (1:n,i) = h * f (x + rk_c (i) * h, yout (1:n))
     end do

     eout (1:n) = (rk_b (1,1) - rk_b (1,2)) * rk_kVectors (1:n,1)

     do i = 2,s
        eout (1:n) = eout (1:n) + (rk_b (i,1) - rk_b (i,2)) * rk_kVectors (1:n,i)
     end do

     hused = h

     eMax = maxval (abs (eout (1:n)) / eBase (1:n)) / eFrac              ! must 1st take abs of eout values
!  write (*,*) ' eMax = ',eMax

     badStep = (eMax > 1.0)

     if (badStep) then
!  write (*,*) ' decrease step, old h = ',h
         h = h * rk_stepSizeSafetyFactor * eMax ** eRescale              ! decrease step size
!  write (*,*) ' decrease step, new h = ',h
     else
         yout (1:n) = y (1:n) + rk_b (1,1) * rk_kVectors (1:n,1)         ! add advancement vector

         do i = 2,s
            yout (1:n) = yout (1:n) + rk_b (i,1) * rk_kVectors (1:n,i)   ! add advancement vector
         end do

!  write (*,'(a,es30.16)') ' x    = ',yout(1)
!  write (*,'(a,es30.16)') ' xmin = ',ymin (1, yout (1:n))
!  write (*,'(a,es30.16)') ' xmax = ',ymax (1, yout (1:n))

         if (any (yout (1:nc) < ymin (nc, yout (1:n)) .or. &             ! check, if RK endpoints
                  yout (1:nc) > ymax (nc, yout (1:n)))     ) then        ! step outside the confinement
             h = h * rk_stepSizeConfinementFactor                        ! barriers
             badStep = .true.
!  write (*,*) ' out of confinement, new h = ',h
             cycle stepSizeLoop
         end if

!  write (*,*) ' increase step, old h = ',h
         h = h * rk_stepSizeSafetyFactor * eMax ** eRescale              ! increase step size
!  write (*,*) ' increase step, new h = ',h

     end if

     hnext = h

  end do stepSizeLoop
!
!
!   ...Alert, if complete non-advancement of dependent variables. Note, that this check
!      has to be done AFTER we evaluated the new dependent variables. There is no
!      point in checking, separately, if the advancement vector for y (1:n) is equal
!      to zero, as the advancement vector can have tiny elements which get lost during
!      addition to y (1:n) due to finite working precision. 
!
!
  stagnation = (maxval (abs (y (1:n) - yout (1:n))) == 0.0)

  if (stagnation) then
      write (*,*) ' h = ',h
      write (*,*) ' y(1) = ',y(1)
      write (*,*) ' y(2) = ',y(2)
      write (*,*) ' y(3) = ',y(3)
      write (*,*) ' yout(1) = ',yout(1)
      write (*,*) ' yout(2) = ',yout(2)
      write (*,*) ' yout(3) = ',yout(3)
      call Driver_abortFlash ('[rk_stepEAC] ERROR: no advancement of dependent variables!')
  end if
!
!
!   ...Ready! 
!
!
  return
end subroutine rk_stepEAC
