!!****if* source/physics/Eos/EosMain/multiTemp/Helmholtz/Ye/eos_helmholtz
!!
!! NAME
!!
!! eos_helmholtz
!!
!! SYNOPSIS
!!
!!  call eos_helmholtz(integer(IN) :: mode,
!!                     integer(IN) :: vecLen,
!!                     real(INOUT) :: eosData(vecLen*EOS_NUM),
!!           optional, real(IN)    :: massFrac(vecLen*NSPECIES),
!!           optional, logical(IN) :: mask(EOS_VARS+1:EOS_NUM)    )
!!
!! DESCRIPTION
!!
!!   Driver for the Helmholtz and Nadyozhin equations of state.
!!   See the NOTES section for important information about this implementation.
!!
!!  This routine applies the equation of state to thermodynamic 
!!  quantities at one or more grid cells.  The number of cells is 
!!  determined by the argument veclen.  Data is packaged for this 
!!  routine in the 1d array, eosData.  The data in eosData is organized: 
!!  1:vecLen points contain the first variable, vecLen+1:2*vecLen points 
!!  contain the second variable, and so on. The number and order of
!!  variables in the array is determined by the constants defined in Eos.h.
!!  
!!  The routine takes different quantities as givens depending on the
!!  value of the mode variable: if mode=MODE_DENS_TEMP, density and
!!  temperature are taken as given, and pressure and internal energy are generated
!!  as output; if mode=MODE_DENS_EI, density and internal energy are taken as
!!  givens, and pressure and temperature are generated as output.  If
!!  mode=MODE_DENS_PRES, density and pressure are taken as givens, and
!!  energy and temperature are generated as output.  Note that internal
!!  energy is grid variable EINT_VAR, not ENER_VAR.
!!  
!!  In addition to pressure, temperature, and internal energy, which are
!!  always thermodynamically consistent after this call, other quantities
!!  such as the various thermodynamic partial derivatives can be
!!  calculated based on the values in the argument, mask.  mask is a
!!  logical array with one entry per quantity, with the order determined
!!  by constants defined in Eos.h (the same as those for the eosData
!!  argument); .true. means return the quantity, .false. means don't.
!!
!!  ARGUMENTS
!! 
!!  mode :    Selects the mode of operation of the Eos unit.
!!             The valid values are MODE_DENS_EI, MODE_DENS_PRES and  
!!             MODE_DENS_TEMP as decribed above.
!!
!!  vecLen   : number of points for each input variable
!!
!!  eosData  : This array is the data structure through which variable values are 
!!             passed in and out of the Eos routine. The arrays is sized as 
!!             EOS_NUM*vecLen. EOS_NUM, and individual input and output
!!             Eos variables are defined in Eos.h. The array is organizes such that
!!             the first 1:vecLen entries represent the first Eos variable, vecLen+1:
!!             2*vecLen represent the second Eos variable and so on. 
!!
!!  massFrac : Contains the mass fractions of the species included in
!!             the simulation. The array is sized as NSPECIES*vecLen.
!!             For the Ye formulation, this value is never used!
!!
!!  mask     : Mask is a logical array the size of EOS_DERIVS (number
!!              of partial derivatives that can be computed, defined in
!!              Eos.h), where each index represents a specific partial derivative
!!              that can be calculated by the Eos unit. A .true. value in mask 
!!              results in the corresponding derivative being calculated and 
!!              returned. It should preferably be dimensioned as
!!              mask(EOS_VARS+1:EOS_NUM) in the calling routine 
!!              to exactly match the arguments declaration in Eos Unit.
!!             Note that the indexing of mask does not begin at 1, but rather at one past
!!             the number of variables.
!!
!!             The Helmholtz EOS kernel calculation ignores the mask setting and calculates
!!             all derivatives, whether needed or not.  This routine does not return
!!             derivatives if the mask is requested, but the calculation is not speeded up
!!             by setting the mask.
!!
!!
!! PARAMETERS
!!
!!  eos_tol    Controls the accuracy of the Newton Rhapson iterations for MODE_DENS_EI and 
!!             MODE_DENS_PRES.
!!
!! EXAMPLE
!!
!! --- A single-point at a time example, does not calculate derivatives (based on Cellular Simulation)---
!!
!!  #include "constants.h"   ! for MODE_DENS_TEMP
!!  #include "Flash.h"       ! for NSPECIES
!!  #include "Eos.h"         ! for EOS_VAR order
!!
!!  real  :: temp_zone, rho_zone, ptot, eint, gamma
!!  real, dimension(EOS_NUM)  :: eosData
!!  real, dimension(SPECIES_BEGIN:SPECIES_END) ::  massFraction  
!!  integer, dimension(2,MDIM)                 :: blockRange,blockExtent
!!
!!
!!  massFraction(:) = 1.0e-12        
!!  massFraction(C12_SPEC) = 1.0
!!
!!  .... initiale temp_zone, rho_zone
!!
!!  call Grid_getBlkIndexLimits(blockId,blockRange,blockExtent)
!!  do k = blockRange(LOW,KAXIS), blockRange(HIGH,KAXIS)
!!     do j = blockRange(LOW,JAXIS),blockRange(HIGH,JAXIS)
!!        do i = blockRange(LOW,IAXIS),blockRange(HIGH,IAXIS)
!!
!!           eosData(EOS_TEMP) = temp_zone
!!           eosData(EOS_DENS) = rho_zone
!!
!!           call Eos(MODE_DENS_TEMP,1,eosData,massFraction)
!!
!!           ptot = eosData(EOS_PRES)
!!           eint = eosData(EOS_EINT)
!!           gamma = eosData(EOS_GAMC)
!!           
!!           call Grid_putPointData(blockId,CENTER,TEMP_VAR,EXTERIOR,iPosition,temp_zone)
!!           call Grid_putPointData(blockId,CENTER,DENS_VAR,EXTERIOR,iPosition,rho_zone)
!!           call Grid_putPointData(blockId,CENTER,PRES_VAR,EXTERIOR,iPosition,ptot)
!!           call Grid_putPointData(blockId,CENTER,EINT_VAR,EXTERIOR,iPosition,eint)
!!             if you want ENER_VAR, calculate it from eint and kinetic energy
!!           call Grid_putPointData(blockId,CENTER,GAMC_VAR,EXTERIOR,iPosition,gamma)
!!           call Grid_putPointData(blockId,CENTER,GAME_VAR,EXTERIOR,iPosition,(ptot/(etot*sim_rhoAmbient) + 1.0))
!!
!!         enddo  ! end of k loop
!!     enddo     ! end of j loop
!!  enddo        ! end of i loop
!!
!! ------------------ Row at a time example, with derivates (based on Eos_unitTest) --------
!!
!!  #include "constants.h"   ! for MODE_DENS_TEMP
!!  #include "Flash.h"       ! for NSPECIES, EOS_NUM
!!  #include "Eos.h"         ! for EOS_VAR order
!!  integer veclen, isize, jsize, ksize, i,j,k, e
!!  real, dimension(:), allocatable :: eosData
!!  real, dimension(:), allocatable :: massFrac
!!  logical, dimension (EOS_VARS+1:EOS_NUM) :: mask
!!  real, allocatable, dimension(:,:,:,:) :: derivedVariables
!!  integer,dimension(2,MDIM) :: blkLimits,blkLimitsGC
!!
!!   ! in the Eos_unitTest, this loops over all blocks.... here is a snippet from inside
!!     call Grid_getBlkIndexLimits(blockID,blkLimits,blkLimitsGC)
!!
!!    !  Allocate the necessary arrays for an entire block of data
!!    isize = (blkLimits(HIGH,IAXIS) - blkLimits(LOW,IAXIS) + 1)
!!    jsize = (blkLimits(HIGH,JAXIS) - blkLimits(LOW,JAXIS) + 1)
!!    ksize = (blkLimits(HIGH,KAXIS) - blkLimits(LOW,KAXIS) + 1)
!!    vecLen=isize
!!    allocate(derivedVariables(isize,jsize,ksize,EOS_NUM))
!!    allocate(eosData(vecLen*EOS_NUM))
!!    allocate(massFrac(vecLen*NSPECIES))
!!    mask = .true.
!!
!!    ! indices into the first location for these variables
!!    pres = (EOS_PRES-1)*vecLen
!!    dens = (EOS_DENS-1)*vecLen
!!    temp = (EOS_TEMP-1)*vecLen
!!
!!
!!    call Grid_getBlkPtr(blockID,solnData)
!!    do k = blkLimits(LOW,KAXIS),blkLimits(HIGH,KAXIS)
!!        do j = blkLimits(LOW,JAXIS), blkLimits(HIGH, JAXIS)
!!           do i = 1,vecLen
!!              massFrac((i-1)*NSPECIES+1:i*NSPECIES) = &
!!                   solnData(SPECIES_BEGIN:SPECIES_END,ib+i-1,j,k)
!!           end do
!!
!!           eosData(pres+1:pres+vecLen) =  solnData(PRES_VAR,ib:ie,j,k)
!!           eosData(dens+1:dens+vecLen) =  solnData(DENS_VAR,ib:ie,j,k)
!!           ! Eos Helmholtz needs a good initial estimate of temperature no matter what the mode
!!           eosData(temp+1:temp+vecLen) =  solnData(TEMP_VAR,ib:ie,j,k)
!!
!!           call Eos(MODE_DENS_PRES,vecLen,eosData,massFrac,mask)
!!
!!           do e=EOS_VARS+1,EOS_NUM
!!              m = (e-1)*vecLen
!!              derivedVariables(1:vecLen,j-NGUARD,k-NGUARD,e) =  eosData(m+1:m+vecLen)
!!           end do
!!        end do
!!     end do
!!
!! NOTES
!!
!!  NSPECIES is defined in Flash.h.
!!
!!  EOS_VARS and EOS_NUM  are defined in Eos.h.
!!  Calling funtions should included Eos.h, in order to get the definitions of
!!  Eos-specific constants to be able to populate the eosData and mask arrays.
!!  
!!  MODE_DENS_TEMP, MODE_DENS_EI, MODE_DENS_PRES, etc. are defined in constants.h.
!!
!!  All routines calling this routine should include a 
!!  use Eos_interface 
!!  statement, preferable with "ONLY" attribute e.g.
!!  use Eos_interface, ONLY:  Eos
!!
!!  The Helmholtz equation of state calculations are iterative for any mode other
!!  than MODE_DENS_TEMP.  Therefore, the intial estimates for temperature and density
!!  must be pretty good upon entering Eos with any other MODE_....or the calculations will
!!  not converge.
!!
!!  This algorithm uses a data table helm_table.dat which contains the coefficients for
!!  one of the interpolating algorithms.  Upon first entry to the Eos, a binary version of this
!!  table (helm_table.bdat) is created for speed of access.  This binary file should NOT be
!!  carried across machine architectures or even compilers.
!!
!!  Here USE_EOS_YE is defined, and this routine operates in "light" or Ye mode.  The abar and zbar are
!!  NOT calculated below, and the calling routine MUST fill up eosData(EOS_ABAR,....) and
!!  eosData(EOS_ZBAR,.....) before calling this.  There is NO ERROR CHECKING within this routine
!!  The user is entirely responsible for figuring this out themselves.  Search in your
!!  object directory for all calls to "call Eos" directly -- the "call Eos_wrapped" has been
!!  taken care of for you!
!!
!!  When operating in MODE_DENS_EI, the INPUT energy is updated.  This change of an input parameter
!!     can be overridden by setting the runtime parameter eos_forceConstantInput to true.
!!     Noted below, see comments prefaced with ConstantInput.
!!  Similarly, when operating in MODE_DENS_PRES, the INPUT pressure is updated.  Physicists need
!!     to be aware of this.  Similarly can be overridden with the runtime parameter/
!!
!!  The accuracy can be adjusted with the parameter eos_tol.
!!
!! SEE ALSO
!! 
!!  Eos.h    defines the variables used.
!!  Eos_wrapped  sets up the data structure.
!!
!!
!!*** 


subroutine eos_helmholtz(mode,vecLen,eosData,massFrac,mask)

  use Driver_interface, ONLY : Driver_abortFlash
  use Multispecies_interface, ONLY : Multispecies_getSumInv, &
    Multispecies_getSumFrac
  use Logfile_interface, ONLY:  Logfile_stampMessage
  
  use eos_helmInterface, ONLY : eos_helm

  use Eos_data, ONLY: eos_smallT, eos_tol, eos_maxNewton, &
       eos_combinedTempRule, eos_forceConstantInput, eos_meshMe
  use eos_vecData, ONLY:  tempRow, denRow, etotRow, abarRow, zbarRow, &
       gamcRow, ptotRow, deaRow, dezRow, stotRow, dsdRow, dstRow, &
       detRow, dptRow, dpdRow, dedRow, pelRow, neRow, etaRow, cvRow, cpRow, &
       tempRadRow,tempIonRow,tempEleRow,&
       eCompRow, pCompRow

  implicit none

#include "constants.h"
#include "Flash.h"
#include "Eos.h"
#include "Eos_components.h"

  !     Arguments
  integer, INTENT(in) :: mode, vecLen
  real, INTENT(inout), dimension(vecLen*EOS_NUM) :: eosData
  real, optional,INTENT(in), dimension(vecLen*NSPECIES) :: massFrac
  ! must correspond to dimensions of Eos_wrapped
  logical,optional, dimension(EOS_VARS+1:EOS_NUM),INTENT(in)::mask

  ! This is the variable that is used internally -- set to false unless mask comes in
  logical, dimension(EOS_VARS+1:EOS_NUM) :: maskInternal

  integer :: componentMask(EOSCOMP_NUM_COMPONENTS)

  integer :: i, k
  integer :: vecBegin,vecEnd
  integer :: pres, temp, dens, gamc, eint
  integer :: tempIon,tempEle,tempRad
  integer :: eintIon,eintEle,eintRad
  integer :: presIon,presEle,presRad
  integer :: abar, zbar
  integer :: entr, dst, dsd
  integer :: dpt, dpd, det, ded, dea, dez, pel, ne, eta, c_v, c_p
  real    :: abarInv, zbarFrac

  ! declare some local storage for the results of the Newton iteration
  real,dimension(vecLen)::  ewantRow, tnew, error,pwantRow
  !  local storage for forcedConstantInput -- could be allocatable, but might be so slow
  !  that it's not worth the small storage save.
  real,dimension(vecLen)::  psaveRow, esaveRow
  logical  :: firstCall = .true.

  !      Fill the pipe with the initial temperature, density, and composition.
  !      The composition is parametrized by abar and zbar, which are computed
  !      from the mass fractions xn().

  !  if you are working with electron abundance mass scalars, then you don't
  !  necessarily have to have mass fractions.  

   ! warn user if mask is used. Helmholtz EOS doesn't use it very much
  maskInternal = .false.
  if (present(mask)) then
     maskInternal = mask
     if (firstCall) then
        call Logfile_stampMessage("[EOS Helmholtz] WARNING!  Mask setting does not speed up Eos Helmholtz calls")
        if (eos_meshMe .EQ. MASTER_PE) &
             write(*,*)"[EOS Helmholtz] WARNING!  Mask setting does not speed up Eos Helmholtz calls"
        firstCall = .false.
     end if
  end if

  vecBegin = 1
  vecEnd = vecLen

  ! These integers are indexes into the lowest location in UNK that contain the appropriate variable
  pres = (EOS_PRES-1)*vecLen
  dens = (EOS_DENS-1)*vecLen
  temp = (EOS_TEMP-1)*vecLen
  eint = (EOS_EINT-1)*vecLen
  gamc = (EOS_GAMC-1)*vecLen
  abar = (EOS_ABAR-1)*vecLen   
  zbar = (EOS_ZBAR-1)*vecLen
  entr = (EOS_ENTR-1)*vecLen
  tempIon = (EOS_TEMPION-1)*vecLen
  tempEle = (EOS_TEMPELE-1)*vecLen
  tempRad = (EOS_TEMPRAD-1)*vecLen
  eintIon = (EOS_EINTION-1)*vecLen
  eintEle = (EOS_EINTELE-1)*vecLen
  eintRad = (EOS_EINTRAD-1)*vecLen
  presIon = (EOS_PRESION-1)*vecLen
  presEle = (EOS_PRESELE-1)*vecLen
  presRad = (EOS_PRESRAD-1)*vecLen

  !! For allocatable arrays, set them up now.
#ifndef FIXEDBLOCKSIZE
  call eos_vecAlloc(vecLen)
#endif

  do k = 1, vecLen

     tempRow(k)    = eosData(temp+k)
     denRow(k)     = eosData(dens+k)
     ewantRow(k)   = eosData(eint+k)   ! store desired internal energy for MODE_DENS_EI etc. cases
     pwantRow(k)   = eosData(pres+k)   ! store desired pressure for MODE_DENS_PRES etc. cases
     abarRow(k)    = eosData(abar+k)   ! NOTE user is assumed to know what they're doing!
     zbarRow(k)    = eosData(zbar+k)   ! NOTE compare to normal Helmholtz where we don't assume

  enddo

  !! Save the input parameters if eos_forceConstantInput is defined
  if (eos_forceConstantInput) then
     esaveRow = ewantRow
     psaveRow = pwantRow
  end if



  !==============================================================================

  !      MODE_DENS_TEMP  temperature and density given

  !      Crank the EOS on the pipes filled above, then fill the FLASH arrays
  !      with the thermodynamic quantities returned by the EOS.

  select case (mode)
  case(MODE_DENS_TEMP)

     tempRadRow(1:vecLen) = tempRow(1:vecLen)
     tempIonRow(1:vecLen) = tempRow(1:vecLen)
     tempEleRow(1:vecLen) = tempRow(1:vecLen)

     call eos_helm(1,vecLen,maskInternal)

     eosData(pres+1:pres+vecLen)=ptotRow(1:vecLen)
     eosData(eint+1:eint+vecLen)=etotRow(1:vecLen)
     eosData(gamc+1:gamc+vecLen)=gamcRow(1:vecLen)
     eosData(entr+1:entr+vecLen)=stotRow(1:vecLen)

  case(MODE_DENS_TEMP_EQUI)

     tempRadRow(1:vecLen) = tempRow(1:vecLen)
     tempIonRow(1:vecLen) = tempRow(1:vecLen)
     tempEleRow(1:vecLen) = tempRow(1:vecLen)

     call eos_helm(1,vecLen,maskInternal)

     eosData(tempIon+1:tempIon+vecLen)= tempRow(1:vecLen)
     eosData(tempEle+1:tempEle+vecLen)= tempRow(1:vecLen)
     eosData(tempRad+1:tempRad+vecLen)= tempRow(1:vecLen)
     eosData(pres+1:pres+vecLen)      = ptotRow(1:vecLen)
     if(maskInternal(EOS_PRESION)) eosData(presIon+1:presIon+vecLen)= pCompRow(EOSCOMP_ION,1:vecLen)
     if(maskInternal(EOS_PRESELE)) eosData(presEle+1:presEle+vecLen)= pCompRow(EOSCOMP_ELE,1:vecLen)
     if(maskInternal(EOS_PRESRAD)) eosData(presRad+1:presRad+vecLen)= pCompRow(EOSCOMP_RAD,1:vecLen)
     eosData(eint+1:eint+vecLen)      = etotRow(1:vecLen)
     if(maskInternal(EOS_EINTION)) eosData(eintIon+1:eintIon+vecLen)= eCompRow(EOSCOMP_ION,1:vecLen)
     if(maskInternal(EOS_EINTELE)) eosData(eintEle+1:eintEle+vecLen)= eCompRow(EOSCOMP_ELE,1:vecLen)
     if(maskInternal(EOS_EINTRAD)) eosData(eintRad+1:eintRad+vecLen)= eCompRow(EOSCOMP_RAD,1:vecLen)
     eosData(gamc+1:gamc+vecLen)      = gamcRow(1:vecLen)
     eosData(entr+1:entr+vecLen)      = stotRow(1:vecLen)

      if (.NOT. all(tempRow(1:vecLen) > 0)) &
           print*,'eos_nr WARN _EQUI:',tempRow(1:vecLen)

  case(MODE_DENS_TEMP_COMP,MODE_DENS_TEMP_GATHER)

     tempRadRow(1:vecLen) = eosData(tempRad+1:tempRad+vecLen)
     tempIonRow(1:vecLen) = eosData(tempIon+1:tempIon+vecLen)
     tempEleRow(1:vecLen) = eosData(tempEle+1:tempEle+vecLen)

     call eos_helm(1,vecLen,maskInternal)

     eosData(pres+1:pres+vecLen)=ptotRow(1:vecLen)
     if(maskInternal(EOS_PRESION)) eosData(presIon+1:presIon+vecLen)=pCompRow(EOSCOMP_ION,1:vecLen)
     if(maskInternal(EOS_PRESELE)) eosData(presEle+1:presEle+vecLen)=pCompRow(EOSCOMP_ELE,1:vecLen)
     if(maskInternal(EOS_PRESRAD)) eosData(presRad+1:presRad+vecLen)=pCompRow(EOSCOMP_RAD,1:vecLen)
     eosData(eint+1:eint+vecLen)=etotRow(1:vecLen)
     if(maskInternal(EOS_EINTION)) eosData(eintIon+1:eintIon+vecLen)=eCompRow(EOSCOMP_ION,1:vecLen)
     if(maskInternal(EOS_EINTELE)) eosData(eintEle+1:eintEle+vecLen)=eCompRow(EOSCOMP_ELE,1:vecLen)
     if(maskInternal(EOS_EINTRAD)) eosData(eintRad+1:eintRad+vecLen)=eCompRow(EOSCOMP_RAD,1:vecLen)
     eosData(gamc+1:gamc+vecLen)=gamcRow(1:vecLen)
     eosData(entr+1:entr+vecLen)=stotRow(1:vecLen)

     if(mode == MODE_DENS_TEMP_GATHER)then
        if(eos_combinedTempRule==0) then
           call eos_newtonRaphson(vecLen, MODE_DENS_EI, .TRUE., maskInternal,cMask=(/1,1,1/))
        endif
     endif

  case(MODE_DENS_TEMP_ION,MODE_DENS_TEMP_ELE,MODE_DENS_TEMP_RAD)

     componentMask(:) = 0
     select case (mode)
     case(MODE_DENS_TEMP_RAD)
        componentMask(EOSCOMP_RAD) = 1
     case(MODE_DENS_TEMP_ELE)
        componentMask(EOSCOMP_ELE) = 1
     case(MODE_DENS_TEMP_ION)
        componentMask(EOSCOMP_ION) = 1
     end select
     tempRadRow(1:vecLen) = eosData(tempRad+1:tempRad+vecLen)
     tempIonRow(1:vecLen) = eosData(tempIon+1:tempIon+vecLen)
     tempEleRow(1:vecLen) = eosData(tempEle+1:tempEle+vecLen)

     call eos_helm(1,vecLen,maskInternal,componentMask)

     if(maskInternal(EOS_PRESION)) eosData(presIon+1:presIon+vecLen)=pCompRow(EOSCOMP_ION,1:vecLen)
     if(maskInternal(EOS_PRESELE)) eosData(presEle+1:presEle+vecLen)=pCompRow(EOSCOMP_ELE,1:vecLen)
     if(maskInternal(EOS_PRESRAD)) eosData(presRad+1:presRad+vecLen)=pCompRow(EOSCOMP_RAD,1:vecLen)
     if(maskInternal(EOS_EINTION)) eosData(eintIon+1:eintIon+vecLen)=eCompRow(EOSCOMP_ION,1:vecLen)
     if(maskInternal(EOS_EINTELE)) eosData(eintEle+1:eintEle+vecLen)=eCompRow(EOSCOMP_ELE,1:vecLen)
     if(maskInternal(EOS_EINTRAD)) eosData(eintRad+1:eintRad+vecLen)=eCompRow(EOSCOMP_RAD,1:vecLen)
     eosData(pres+1:pres+vecLen)=ptotRow(1:vecLen)
     eosData(eint+1:eint+vecLen)=etotRow(1:vecLen)
     eosData(gamc+1:gamc+vecLen)=gamcRow(1:vecLen)
     eosData(entr+1:entr+vecLen)=stotRow(1:vecLen)

     !==============================================================================
     !      MODE_DENS_EI  internal energy and density given

  case(MODE_DENS_EI, MODE_DENS_EI_SCATTER)

     call eos_newtonRaphson(vecLen, MODE_DENS_EI, .TRUE., maskInternal,cMask=(/1,1,1/))

     if (mode==MODE_DENS_EI_SCATTER) then
        eosData(tempIon+1:tempIon+vecLen)=tempRow(1:vecLen)
        eosData(tempEle+1:tempEle+vecLen)=tempRow(1:vecLen)
        eosData(tempRad+1:tempRad+vecLen)=tempRow(1:vecLen)
        if(maskInternal(EOS_PRESION)) eosData(presIon+1:presIon+vecLen)=pCompRow(EOSCOMP_ION,1:vecLen)
        if(maskInternal(EOS_PRESELE)) eosData(presEle+1:presEle+vecLen)=pCompRow(EOSCOMP_ELE,1:vecLen)
        if(maskInternal(EOS_PRESRAD)) eosData(presRad+1:presRad+vecLen)=pCompRow(EOSCOMP_RAD,1:vecLen)
        if(maskInternal(EOS_EINTION)) eosData(eintIon+1:eintIon+vecLen)=eCompRow(EOSCOMP_ION,1:vecLen)
        if(maskInternal(EOS_EINTELE)) eosData(eintEle+1:eintEle+vecLen)=eCompRow(EOSCOMP_ELE,1:vecLen)
        if(maskInternal(EOS_EINTRAD)) eosData(eintRad+1:eintRad+vecLen)=eCompRow(EOSCOMP_RAD,1:vecLen)
     end if


  case(MODE_DENS_EI_ION,MODE_DENS_EI_ELE,MODE_DENS_EI_RAD)

     componentMask(:) = 0
     select case (mode)
     case(MODE_DENS_EI_RAD)
        componentMask(EOSCOMP_RAD) = 1
        ewantRow(1:vecLen) = eosData(eintRad+1:eintRad+vecLen)   ! store desired radiation energy
     case(MODE_DENS_EI_ELE)
        componentMask(EOSCOMP_ELE) = 1
        ewantRow(1:vecLen) = eosData(eintEle+1:eintEle+vecLen)   ! store desired electron energy
     case(MODE_DENS_EI_ION)
        componentMask(EOSCOMP_ION) = 1
        ewantRow(1:vecLen) = eosData(eintIon+1:eintIon+vecLen)   ! store desired ion energy
     end select
     call eos_newtonRaphson(vecLen, mode, .TRUE., maskInternal,componentMask)


  case(MODE_DENS_EI_COMP,MODE_DENS_EI_ALL,MODE_DENS_EI_GATHER)

     tempRadRow(1:vecLen) = eosData(tempRad+1:tempRad+vecLen)
     tempIonRow(1:vecLen) = eosData(tempIon+1:tempIon+vecLen)
     tempEleRow(1:vecLen) = eosData(tempEle+1:tempEle+vecLen)

     componentMask(:) = (/0,0,0/)
     componentMask(EOSCOMP_RAD) = 1
     tempRow(1:vecLen) = tempRadRow(1:vecLen)
     ewantRow(1:vecLen) = eosData(eintRad+1:eintRad+vecLen)   ! store desired radiation energy
     call eos_newtonRaphson(vecLen, MODE_DENS_EI_RAD, .TRUE., maskInternal,componentMask)
     eosData(tempRad+1:tempRad+vecLen)=tempRow(1:vecLen)
     if(maskInternal(EOS_PRESRAD)) eosData(presRad+1:presRad+vecLen)=ptotRow(1:vecLen)
     if (eos_forceConstantInput)  then
!!$        eosData(eintRad+1:eintRad+vecLen) = esaveRow(1:vecLen)
     else
        if(maskInternal(EOS_EINTRAD)) eosData(eintRad+1:eintRad+vecLen) = etotRow(1:vecLen)
     end if

     componentMask(:) = (/0,0,0/)
     componentMask(EOSCOMP_ELE) = 1
     tempRow(1:vecLen) = tempEleRow(1:vecLen)
     ewantRow(1:vecLen) = eosData(eintEle+1:eintEle+vecLen)   ! store desired electron energy
     call eos_newtonRaphson(vecLen, MODE_DENS_EI_ELE, .TRUE., maskInternal,componentMask)
     eosData(tempEle+1:tempEle+vecLen)=tempRow(1:vecLen)
     if(maskInternal(EOS_PRESELE)) eosData(presEle+1:presEle+vecLen)=ptotRow(1:vecLen)
     if (eos_forceConstantInput)  then
!!$        eosData(eintEle+1:eintEle+vecLen) = esaveRow(1:vecLen)
     else
        if(maskInternal(EOS_EINTELE)) eosData(eintEle+1:eintEle+vecLen) = etotRow(1:vecLen)
     end if

     componentMask(:) = (/0,0,0/)
     componentMask(EOSCOMP_ION) = 1
     tempRow(1:vecLen) = tempIonRow(1:vecLen)
     ewantRow(1:vecLen) = eosData(eintIon+1:eintIon+vecLen)   ! store desired ion energy
     call eos_newtonRaphson(vecLen, MODE_DENS_EI_ION, .TRUE., maskInternal,componentMask)
     eosData(tempIon+1:tempIon+vecLen)=tempRow(1:vecLen)
     if(maskInternal(EOS_PRESION)) eosData(presIon+1:presIon+vecLen)=ptotRow(1:vecLen)
     if (eos_forceConstantInput)  then
!!$        eosData(eintIon+1:eintIon+vecLen) = esaveRow(1:vecLen)
     else
        if(maskInternal(EOS_EINTION)) eosData(eintIon+1:eintIon+vecLen) = etotRow(1:vecLen)
     end if

     if (mode==MODE_DENS_EI_ALL) then
        componentMask(:) = (/1,1,1/)
        tempRow(1:vecLen) = eosData(temp+1:temp+vecLen)
        ewantRow(1:vecLen) = eosData(eint+1:eint+vecLen)   ! re-store desired internal energy
        call eos_newtonRaphson(vecLen, MODE_DENS_EI, .TRUE., maskInternal,componentMask)
     end if
     if (mode==MODE_DENS_EI_GATHER) then
        tempRow(1:vecLen) = eosData(temp+1:temp+vecLen)
        if (ANY(maskInternal((/EOS_PRESELE,EOS_PRESION,EOS_PRESRAD/)) .EQV. .FALSE.)) then
           call Driver_abortFlash("[Eos] cannot calculate MODE_DENS_EI_GATHER without component pressure masks.&
                & Set mask appropriately.")
        end if
        ewantRow(1:vecLen) = eosData(eintIon+1:eintIon+vecLen)+ &
                             eosData(eintEle+1:eintEle+vecLen)+ &
                             eosData(eintRad+1:eintRad+vecLen)
        call eos_newtonRaphson(vecLen, MODE_DENS_EI, .TRUE., maskInternal,cMask=(/1,1,1/))
     end if


     !==============================================================================

     !      MODE_DENS_PRES  pressure and density given

  case(MODE_DENS_PRES)

     componentMask(:) = 1
     call eos_newtonRaphson(vecLen, mode, .FALSE., maskInternal, componentMask)


     !==============================================================================

     ! Unknown EOS mode selected

  case default
     if (eos_meshMe .EQ. MASTER_PE) print*, '[eos_helmholtz] Error: unknown input mode', mode
     call Driver_abortFlash('[Eos] Error: unknown input mode in subroutine eos_helmholtz')
  end select


  ! Get the optional values
  if(present(mask)) then
     ! Entropy derivatives
     if(mask(EOS_DST)) then
        dst = (EOS_DST-1)*vecLen
        eosData(dst+1:dst+vecLen) = dstRow(1:vecLen)
     end if
     if(mask(EOS_DSD)) then
        dsd = (EOS_DSD-1)*vecLen
        eosData(dsd+1:dsd+vecLen) = dsdRow(1:vecLen)
     end if
     if(mask(EOS_DPT)) then
        dpt = (EOS_DPT-1)*vecLen
         eosData(dpt+1:dpt+vecLen) = dptRow(1:vecLen)
     end if
     if(mask(EOS_DPD)) then
        dpd = (EOS_DPD-1)*vecLen
        eosData(dpd+1:dpd+vecLen) = dpdRow(1:vecLen)
     end if
     if(mask(EOS_DET))then
        det = (EOS_DET-1)*vecLen
         eosData(det+1:det+vecLen) = detRow(1:vecLen)
     end if
     if(mask(EOS_DED))then 
        ded = (EOS_DED-1)*vecLen
        eosData(ded+1:ded+vecLen) = dedRow(1:vecLen)
     end if
     if(mask(EOS_DEA))then 
        dea = (EOS_DEA-1)*vecLen
        eosData(dea+1:dea+vecLen) = deaRow(1:vecLen)
     end if
     if(mask(EOS_DEZ))then 
        dez = (EOS_DEZ-1)*vecLen
        eosData(dez+1:dez+vecLen) = dezRow(1:vecLen)
     end if
     if(mask(EOS_PEL))then 
        pel = (EOS_PEL-1)*vecLen
        eosData(pel+1:pel+vecLen) = pelRow(1:vecLen)
     end if
     if(mask(EOS_NE))then 
        ne = (EOS_NE-1)*vecLen
        eosData(ne+1:ne+vecLen) = neRow(1:vecLen)
     end if
     if(mask(EOS_ETA))then 
        eta = (EOS_ETA-1)*vecLen
        eosData(eta+1:eta+vecLen) = etaRow(1:vecLen)
     end if
     
     if(mask(EOS_CV))then
        if(mask(EOS_DET)) then
           c_v = (EOS_CV-1)*vecLen
           eosData(c_v+1:c_v+vecLen) = cvRow(1:vecLen)
        else
           call Driver_abortFlash("[eos_helmholtz] cannot calculate C_V without DET.  Set mask appropriately.")
        end if
     end if
     
     if(mask(EOS_CP))then
        if(mask(EOS_CV).and.mask(EOS_DET)) then
           c_p = (EOS_CP-1)*vecLen
           eosData(c_p+1:c_p+vecLen) = cpRow(1:vecLen)
        else
           call Driver_abortFlash("[eos_helmholtz] cannot calculate C_P without C_V and DET.  Set mask appropriately.")
        end if
     end if
  end if

  !! DEV: BEWARE CHEATING FOR NOW , THINKING FOR LATER

    if(mask(EOS_CVELE))then
        if(mask(EOS_DET)) then
           c_v = (EOS_CVELE-1)*vecLen
           eosData(c_v+1:c_v+vecLen) = cvRow(1:vecLen)
        else
           call Driver_abortFlash("[eos_helmholtz] cannot calculate C_VELE without DET.  Set mask appropriately.")
        end if
     end if
  !!




  !! Close up arrays if previously allocated
#ifndef FIXEDBLOCKSIZE  
  call eos_vecDealloc()
#endif

  return

  contains
    subroutine eos_newtonRaphson(vecLen, mode, enerWanted, eosMask,cMask)
      implicit none
      integer,intent(IN) :: vecLen, mode
      logical,intent(IN) :: enerWanted
!!$  real,dimension(vecLen),intent(IN)    ::  XXwantRow
!!$  real,dimension(vecLen),intent(INOUT) ::  XXtotRow
!!$  real,dimension(vecLen),intent(OUT)   ::  dXXtRow
      logical,optional, dimension(EOS_VARS+1:EOS_NUM),INTENT(in)::eosMask
      integer,optional, dimension(EOSCOMP_NUM_COMPONENTS),INTENT(in)::cMask


      ! Initialize the errors
      error(:) = 0.0e0

      ! Do the first eos call with all the zones in the pipe
      !  NOTE that eos_helm can ONLY operate in the equivalent of
      !  MODE_DENS_TEMP, as it returns pressure, energy and derivatives only
      !  So if you send in a crappy temperature here, you'll get a crappy starting
      !  position and the iteration won't converge.
      !  Initial temperature here is what is stored in the grid, even though we 
      !    SUSPECT this is not in equilibrium (or we wouldn't be calling Eos if it was fine)
      if (.NOT. all(tempRow(1:vecLen) > 0)) &
           print*,'eos_nr WARN:',tempRow(1:vecLen)

      if (.NOT. present(cMask)) then
         tempRadRow(1:vecLen) = tempRow(1:vecLen)
         tempIonRow(1:vecLen) = tempRow(1:vecLen)
         tempEleRow(1:vecLen) = tempRow(1:vecLen)
      else
         if (cMask(EOSCOMP_RAD).NE.0) tempRadRow(1:vecLen) = tempRow(1:vecLen)
         if (cMask(EOSCOMP_ION).NE.0) tempIonRow(1:vecLen) = tempRow(1:vecLen)
         if (cMask(EOSCOMP_ELE).NE.0) tempEleRow(1:vecLen) = tempRow(1:vecLen)
      end if
      call eos_helm(vecBegin,vecEnd,eosMask,componentMask=cMask)
      !  Now eos_helm has returned ptotRow, etotRow, dXXtRow, and gamcRow


      !  Create initial condition
      do k = vecBegin, vecEnd

         if (enerWanted) then
            !  ewantRow is our desired EI input
            tnew(k) = tempRow(k) - (etotRow(k) - ewantRow(k)) / detRow(k)
         else                    !pressure is wanted
            tnew(k) = tempRow(k) - (ptotRow(k) - pwantRow(k)) / dptRow(k)
         end if

         ! Don't allow the temperature to change by more than an order of magnitude 
         ! in a single iteration
         if (tnew(k) .GT. 10.e0*tempRow(k)) tnew(k) =  & 
              &           10.e0*tempRow(k)
         if (tnew(k) .LT. 0.1e0*tempRow(k)) tnew(k) =  & 
              &           0.1e0*tempRow(k)

         ! Compute the error
         error(k) = abs((tnew(k) - tempRow(k)) / tempRow(k))

         ! Store the new temperature
         tempRow(k) = tnew(k)

         ! Check if we are freezing, if so set the temperature to smallt, and adjust 
         ! the error so we don't wait for this one
         if (tempRow(k) .LT. eos_smallt) then
            tempRow(k) = eos_smallt
            error(k)    = 0.1*eos_tol
         endif

      enddo

      ! Loop over the zones individually now
      do k = vecBegin, vecEnd
         do i = 2, eos_maxNewton
            if (error(k) .LT. eos_tol) goto 70

            if (.NOT. present(cMask)) then
               tempRadRow(k) = tempRow(k)
               tempIonRow(k) = tempRow(k)
               tempEleRow(k) = tempRow(k)
            else
               if (cMask(EOSCOMP_RAD).NE.0) tempRadRow(k) = tempRow(k)
               if (cMask(EOSCOMP_ION).NE.0) tempIonRow(k) = tempRow(k)
               if (cMask(EOSCOMP_ELE).NE.0) tempEleRow(k) = tempRow(k)
            end if
            ! do eos only over this single item
            call eos_helm(k,k,eosMask,componentMask=cMask)

            if (enerWanted) then
               tnew(k) = tempRow(k) - (etotRow(k) - ewantRow(k)) / detRow(k)
            else                    !pressure is wanted
               tnew(k) = tempRow(k) - (ptotRow(k) - pwantRow(k)) / dptRow(k)
            end if

            ! Don't allow the temperature to change by more than an order of magnitude 
            ! in a single iteration
            if (tnew(k) .GT. 10.e0*tempRow(k)) tnew(k) =  & 
                 &              10.e0*tempRow(k)
            if (tnew(k) .LT. 0.1e0*tempRow(k)) tnew(k) =  & 
                 &              0.1e0*tempRow(k)

            ! Compute the error
            error(k) = abs((tnew(k) - tempRow(k)) / tempRow(k))

            ! Store the new temperature
            tempRow(k) = tnew(k)

            ! Check if we are freezing, if so set the temperature to eos_smallt, and adjust 
            ! the error so we don't wait for this one
            if (tempRow(k) .LT. eos_smallt) then
               tempRow(k) = eos_smallt
               error(k)    = .1*eos_tol
            endif

         end do  ! end of Newton iterations loop.  Failure drops below, success goes to 70

         ! Land here if too many iterations are needed -- failure

         print *, ' '
         print *, 'Newton-Raphson failed in routine Helmholtz Eos'

         select case(mode)
         case(MODE_DENS_EI)
            print *, '(e and rho as input):'
         case(MODE_DENS_PRES)
            print *, '(pres and rho as input):'
         case default
            print *, 'Unusual Eos mode:', mode
         end select
         print *, ' '
         print *, 'too many iterations'
         print *, ' '
         print *, ' k    = ', k,vecBegin,vecEnd
         print *, ' temp = ', tempRow(k)
         print *, ' dens = ', denRow(k)
         print *, ' pres = ', ptotRow(k)

         call Driver_abortFlash('[Eos] Error: too many iteration in Newton-Raphson')


         ! Land here if the Newton iteration converged
         !  jumps out of the iterations, but then continues to the next vector location

70       continue           
      end do

      ! Crank through the entire eos one last time

      if (.NOT. present(cMask)) then
         tempRadRow(1:vecLen) = tempRow(1:vecLen)
         tempIonRow(1:vecLen) = tempRow(1:vecLen)
         tempEleRow(1:vecLen) = tempRow(1:vecLen)
      else
         if (cMask(EOSCOMP_RAD).NE.0) tempRadRow(1:vecLen) = tempRow(1:vecLen)
         if (cMask(EOSCOMP_ION).NE.0) tempIonRow(1:vecLen) = tempRow(1:vecLen)
         if (cMask(EOSCOMP_ELE).NE.0) tempEleRow(1:vecLen) = tempRow(1:vecLen)
      end if
      call eos_helm(vecBegin,vecEnd,eosMask,componentMask=cMask)

      ! Fill the FLASH arrays with the results.  
      !  In MODE_DENS_EI, we should be generating temperature and pressure (plus gamma and entropy)

      select case(mode)
      case(MODE_DENS_EI)
         eosData(temp+1:temp+vecLen)=tempRow(1:vecLen)
         eosData(pres+1:pres+vecLen)=ptotRow(1:vecLen)
         !  Update the energy to be the true energy, instead of the energy we were trying to meet
         !  ConstantInput LBR and KW believe this is WRONG -- the input arrays should not be changed
         if (eos_forceConstantInput)  then
            eosData(eint+1:eint+vecLen) = esaveRow(1:vecLen)
         else
            eosData(eint+1:eint+vecLen) = etotRow(1:vecLen)
         end if
      case(MODE_DENS_PRES)
         eosData(temp+1:temp+vecLen)=tempRow(1:vecLen)
         ! Update the pressure to be the equilibrium pressure, instead of the pressure we were trying to meet
         !  ConstantInput LBR and KW believe this is wrong.  See notes at the top of the routine
         if (eos_forceConstantInput) then
            eosData(pres+1:pres+vecLen) = psaveRow(1:vecLen)
         else
            eosData(pres+1:pres+vecLen) = ptotRow(1:vecLen)
         end if
         eosData(eint+1:eint+vecLen)=etotRow(1:vecLen)
      end select

      eosData(gamc+1:gamc+vecLen)=gamcRow(1:vecLen)
      eosData(entr+1:entr+vecLen)=stotRow(1:vecLen)

    end subroutine eos_newtonRaphson


  end subroutine eos_helmholtz


