!!****if* source/physics/Hydro/HydroMain/split/RHD/Hydro
!!
!! NAME
!!
!!  Hydro
!!
!! SYNOPSIS
!!
!!  Hydro( integer (IN):: blockCount, 
!!         integer (IN):: blockList(blockCount), 
!!         real    (IN):: timeEndAdv, 
!!         real    (IN):: dt, 
!!         real    (IN):: dtOld, 
!!         integer (IN):: sweepOrder )
!!
!! DESCRIPTION
!!   
!!  Perform a 1, 2, or 3D hydro update.  This version handles
!!  directionally split hydro schemes.  The input
!!  parameter sweepOrder determines the ordering of sweep
!!  directions.  For example, in 3d, SWEEP_XYZ means to perform 
!!  1d sweeps first in the x direction, then the y, then z direction, 
!!  while SWEEP_ZYX means to performs the 1d sweeps in the opposite order.
!!
!!  The blockList and blockCount arguments tell this routine on 
!!  which blocks and on how many to operate.  blockList is an 
!!  integer array of size blockCount that contains the local 
!!  block numbers of blocks on which to advance.
!!
!!  dt gives the timestep through which this update should advance,
!!  and timeEndAdv tells the time that this update will reach when
!!  it finishes.  dtOld gives the previously taken timestep.  
!!  
!! ARGUMENTS
!!
!!  blockCount -  the number of blocks in blockList
!!  blockList -   array holding local IDs of blocks on which to advance
!!  timeEndAdv -  end time
!!  dt -          timestep
!!  dtOld -       old timestep
!!  sweepOrder -  direction of hydro sweep, can be: SWEEP_XYZ or SWEEP_ZYX
!!                as defined in  constants.h
!!
!! NOTES
!!
!!  PLUTO would be a better choice for doing Relativistic Hydro simulations than FLASH.
!!
!!***

subroutine Hydro(         &
                   blockCount, blockList, &
                   timeEndAdv, dt, dtOld, &
                   sweepOrder )

  use Hydro_data,       ONLY:  hy_dtmin
  use hy_rhd_interface, ONLY : hy_rhd_sweep

  implicit none

#include "constants.h"
#include "Flash.h"

  integer, INTENT(IN) :: blockCount
  integer, INTENT(IN), dimension(blockCount) :: blockList
  real,    INTENT(IN) :: timeEndAdv, dt, dtOld
  integer, INTENT(IN) :: sweepOrder

  !!--------------------------------------------------------------------------
  !!  Perform three sets of one-dimensional hydro sweeps
  !!--------------------------------------------------------------------------

!!$  rhd_dtmin=1.0e99
  select case (sweepOrder)

  case (SWEEP_XYZ)
     call hy_rhd_sweep ( blockCount, blockList, &
                         timeEndAdv, dt, dtOld, &
                         SWEEP_X )

     if (NDIM >= 2) call  hy_rhd_sweep( blockCount, blockList, &
                                        timeEndAdv, dt, dtOld, &
                                        SWEEP_Y )
     
     if (NDIM == 3) call  hy_rhd_sweep( blockCount, blockList, &
                                        timeEndAdv, dt, dtOld, &
                                        SWEEP_Z )
  case (SWEEP_XZY)

     call hy_rhd_sweep ( blockCount, blockList, &
                         timeEndAdv, dt, dtOld,  &
                         SWEEP_X )

     if (NDIM == 3) call  hy_rhd_sweep( blockCount, blockList, &
                                        timeEndAdv, dt, dtOld, &
                                        SWEEP_Z )

     if (NDIM >= 2) call  hy_rhd_sweep( blockCount, blockList, &
                                        timeEndAdv, dt, dtOld, &
                                        SWEEP_Y )

  case (SWEEP_ZYX)
     
     if (NDIM == 3) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Z )

     if (NDIM >= 2) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Y )

     call hy_rhd_sweep( blockCount, blockList, &
                        timeEndAdv, dt, dtOld, &
                        SWEEP_X )

  case (SWEEP_ZXY)
     
     if (NDIM == 3) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Z )

     call hy_rhd_sweep( blockCount, blockList, &
                        timeEndAdv, dt, dtOld, &
                        SWEEP_X )

     if (NDIM >= 2) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Y )

  case (SWEEP_YXZ)
     

     if (NDIM >= 2) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Y )

     call hy_rhd_sweep( blockCount, blockList, &
                        timeEndAdv, dt, dtOld, &
                        SWEEP_X )

     if (NDIM == 3) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Z )
  case (SWEEP_YZX)
     
     if (NDIM >= 2) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Y )

     if (NDIM == 3) call hy_rhd_sweep( blockCount, blockList, &
                                       timeEndAdv, dt, dtOld, &
                                       SWEEP_Z )

     call hy_rhd_sweep( blockCount, blockList, &
                        timeEndAdv, dt, dtOld, &
                        SWEEP_X )

  end select


end subroutine Hydro
