!!****if* source/physics/sourceTerms/Heatexchange/HeatexchangeMain/Spitzer/hx_ieEquilTime
!!
!! NAME
!!
!!  hx_ieEquilTime
!!
!! SYNOPSIS
!!
!!  call hx_ieEquilTime(real, intent(in )  :: zbar,
!!                      real, intent(in )  :: abar,
!!                      real, intent(in )  :: tele,
!!                      real, intent(in )  :: tion,
!!                      real, intent(in )  :: nion,
!!                      real(out) :: eqtime)
!!
!! DESCRIPTION
!!
!! Compute the Coulomb Logarithm using the form presented in:
!!     H. Brysk, P.M. Campbell, P. Hammerling, 
!!     "Thermal Conduction in Laser Fusion", 
!!     Plasma Physics, vol. 17, pp. 473, (1974). 
!!
!! ARGUMENTS
!!
!!   zbar : Average ionization
!!   abar : Average atomic mass
!!   tele : electron temperature
!!   tion : ion temperature
!!   nion : electron number density
!!   eqtime : ion/electron equilibration time
!!
!!
!!***

subroutine hx_ieEquilTime(zbar, abar, tele, tion, nion, eqtime)
  use Heatexchange_data,ONLY: hx_ieTimeCoef, hx_boltz, hx_qele, &
       hx_navo, hx_mele, hx_hbar
  implicit none

#include "constants.h"

  ! Arguments:
  real, intent(in ) :: zbar
  real, intent(in ) :: abar
  real, intent(in ) :: tele
  real, intent(in ) :: tion
  real, intent(in ) :: nion
  real, intent(out) :: eqtime

  ! Local variables:
  real :: ll
  real :: mion

  call loglambda(tele, zbar*nion, zbar, ll)
  mion = abar / hx_navo

  eqtime = hx_ieTimeCoef * &
       3.0 * hx_boltz**1.5 / (8.0 * sqrt(2*PI) * hx_qele**4) * &
       (mion * tele + hx_mele * tion)**1.5 / &
       ( sqrt(mion*hx_mele) * zbar**2 * nion * ll)

contains

  subroutine loglambda(tele, nele, zbar, ll)
    implicit none

    ! This subroutine computes the Coulomb logarithm. The formula used
    ! comes from Atzeni.
    
    real, intent(in)  :: tele ! electron temperature [K]
    real, intent(in)  :: nele ! electron number density [cm^-3]
    real, intent(in)  :: zbar ! the average ionization [unitless]
    real, intent(out) :: ll   ! the coulomb logarithm [unitless]

    real :: bmax, bmin, bmin_classic, bmin_quantum
    real, parameter :: ll_floor = 1.0
    
    bmax = sqrt(hx_boltz * tele / (4*PI * hx_qele**2 * nele))
    
    bmin_classic = zbar * hx_qele**2 / (3*hx_boltz*tele)
    bmin_quantum = hx_hbar / (2*sqrt(3*hx_boltz*tele*hx_mele))
    bmin = max(bmin_classic, bmin_quantum)

    ll = max(log(bmax/bmin), ll_floor)

  end subroutine loglambda


end subroutine hx_ieEquilTime
