!*******************************************************************************

! Routine:      amr_prolong_gen_work_fun

! Description:

! This routine takes data from the array recv1, originally extracted
! from the workspace array work on some block, and performs a prolongation
! operation on it, between the bounds ranges ia to ib, ja to jb, and ka to kb.
! The data in recv1 is from a parent block and the result of the prolongation
! operation is returned in outp.  The position of the child within the parent
! block is specified by the ioff, joff and koff arguments.

! This particular prolongation is conservative, cell-averaged, triquartic
! interpolation. It can be used for blocks with an even or odd number of grid
! cells.


      subroutine amr_prolong_gen_work_fun & 
     &       (ia, ib, ja, jb, ka, kb, lb, ioff, joff, koff, mype)

!===============================================================================

      use runtime_parameters
use physicaldata
      use tree
      use workspace
      implicit none
      include 'mpif.h'



      integer :: ia, ib, ja, jb, ka, kb, lb
      integer :: ioff, joff, koff, mype

      integer :: ip, ip1, im1, jp, jp1, jm1, kp, kp1, km1
      integer :: i, j, k, ico, jco, kco
      integer :: ic, jc, kc, icmin, icmax, jcmin, jcmax
      integer :: kcmin, kcmax
      real    :: cmin, cmax, fmin, fmax

      real, save    :: PC(-2:2,0:1)
      logical, save :: first_call = .true., monotone

!===============================================================================

      if (first_call) then
        call get_parm_from_context & 
     &    (global_parm_context, "monotone", monotone)
        PC(:,1) = (/ -3./128.,  11./64., 1., -11./64.,  3./128. /)
        PC(:,0) = (/  3./128., -11./64., 1.,  11./64., -3./128. /)
        first_call = .false.
      endif

! Interpolation loop:  loop over zones in the target (fine, child) block.
! (i,j,k) are the indices of the fine zones.  (ip,jp,kp) are the indices
! of the coarse zone enclosing each fine zone.  ip1 and im1 refer to coarse
! zones offset by one to the right and left, respectively; likewise for
! jp1, jm1, kp1, km1.  ico, jco, and kco indicate which set of coefficients
! to use.

      do k = ka, kb

        kp  = ((k-1)/2) + 1 + (nguard_work/2)*k3d + koff
        kp1 = kp + k3d
        km1 = kp - k3d
        kco = mod(k, 2)

        do j = ja, jb

          jp  = ((j-1)/2) + 1 + (nguard_work/2)*k2d + joff
          jp1 = jp + k2d
          jm1 = jp - k2d
          jco = mod(j, 2)

          do i = ia, ib

            ip  = ((i-1)/2) + 1 + (nguard_work/2) + ioff
            ip1 = ip + 1
            im1 = ip - 1
            ico = mod(i, 2)

#if N_DIM == 1
            work(i,j,k,lb) = &
              PC(-2,ico) * recv1(ip-2,jp,kp) + &
              PC(-1,ico) * recv1(ip-1,jp,kp) + &
              PC( 0,ico) * recv1(ip,  jp,kp) + &
              PC( 1,ico) * recv1(ip+1,jp,kp) + &
              PC( 2,ico) * recv1(ip+2,jp,kp)
#endif
#if N_DIM == 2 
            work(i,j,k,lb) = &
              PC(-2,jco) * (PC(-2,ico) * recv1(ip-2,jp-2,kp) + &
                            PC(-1,ico) * recv1(ip-1,jp-2,kp) + &
                            PC( 0,ico) * recv1(ip,  jp-2,kp) + &
                            PC( 1,ico) * recv1(ip+1,jp-2,kp) + &
                            PC( 2,ico) * recv1(ip+2,jp-2,kp)) + &
              PC(-1,jco) * (PC(-2,ico) * recv1(ip-2,jp-1,kp) + &
                            PC(-1,ico) * recv1(ip-1,jp-1,kp) + &
                            PC( 0,ico) * recv1(ip,  jp-1,kp) + &
                            PC( 1,ico) * recv1(ip+1,jp-1,kp) + &
                            PC( 2,ico) * recv1(ip+2,jp-1,kp)) + &
              PC( 0,jco) * (PC(-2,ico) * recv1(ip-2,jp,  kp) + &
                            PC(-1,ico) * recv1(ip-1,jp,  kp) + &
                            PC( 0,ico) * recv1(ip,  jp,  kp) + &
                            PC( 1,ico) * recv1(ip+1,jp,  kp) + &
                            PC( 2,ico) * recv1(ip+2,jp,  kp)) + &
              PC( 1,jco) * (PC(-2,ico) * recv1(ip-2,jp+1,kp) + &
                            PC(-1,ico) * recv1(ip-1,jp+1,kp) + &
                            PC( 0,ico) * recv1(ip,  jp+1,kp) + &
                            PC( 1,ico) * recv1(ip+1,jp+1,kp) + &
                            PC( 2,ico) * recv1(ip+2,jp+1,kp)) + &
              PC( 2,jco) * (PC(-2,ico) * recv1(ip-2,jp+2,kp) + &
                            PC(-1,ico) * recv1(ip-1,jp+2,kp) + &
                            PC( 0,ico) * recv1(ip,  jp+2,kp) + &
                            PC( 1,ico) * recv1(ip+1,jp+2,kp) + &
                            PC( 2,ico) * recv1(ip+2,jp+2,kp))
#endif
#if N_DIM == 3  
            work(i,j,k,lb) = &
              PC(-2,kco) * (PC(-2,jco) * (PC(-2,ico) * recv1(ip-2,jp-2,kp-2) + &
                                          PC(-1,ico) * recv1(ip-1,jp-2,kp-2) + &
                                          PC( 0,ico) * recv1(ip,  jp-2,kp-2) + &
                                          PC( 1,ico) * recv1(ip+1,jp-2,kp-2) + &
                                          PC( 2,ico) * recv1(ip+2,jp-2,kp-2))+ &
                            PC(-1,jco) * (PC(-2,ico) * recv1(ip-2,jp-1,kp-2) + &
                                          PC(-1,ico) * recv1(ip-1,jp-1,kp-2) + &
                                          PC( 0,ico) * recv1(ip,  jp-1,kp-2) + &
                                          PC( 1,ico) * recv1(ip+1,jp-1,kp-2) + &
                                          PC( 2,ico) * recv1(ip+2,jp-1,kp-2))+ &
                            PC( 0,jco) * (PC(-2,ico) * recv1(ip-2,jp,  kp-2) + &
                                          PC(-1,ico) * recv1(ip-1,jp,  kp-2) + &
                                          PC( 0,ico) * recv1(ip,  jp,  kp-2) + &
                                          PC( 1,ico) * recv1(ip+1,jp,  kp-2) + &
                                          PC( 2,ico) * recv1(ip+2,jp,  kp-2))+ &
                            PC( 1,jco) * (PC(-2,ico) * recv1(ip-2,jp+1,kp-2) + &
                                          PC(-1,ico) * recv1(ip-1,jp+1,kp-2) + &
                                          PC( 0,ico) * recv1(ip,  jp+1,kp-2) + &
                                          PC( 1,ico) * recv1(ip+1,jp+1,kp-2) + &
                                          PC( 2,ico) * recv1(ip+2,jp+1,kp-2))+ &
                            PC( 2,jco) * (PC(-2,ico) * recv1(ip-2,jp+2,kp-2) + &
                                          PC(-1,ico) * recv1(ip-1,jp+2,kp-2) + &
                                          PC( 0,ico) * recv1(ip,  jp+2,kp-2) + &
                                          PC( 1,ico) * recv1(ip+1,jp+2,kp-2) + &
                                          PC( 2,ico) * recv1(ip+2,jp+2,kp-2)))
            work(i,j,k,lb) = work(i,j,k,lb) + &
              PC(-1,kco) * (PC(-2,jco) * (PC(-2,ico) * recv1(ip-2,jp-2,kp-1) + &
                                          PC(-1,ico) * recv1(ip-1,jp-2,kp-1) + &
                                          PC( 0,ico) * recv1(ip,  jp-2,kp-1) + &
                                          PC( 1,ico) * recv1(ip+1,jp-2,kp-1) + &
                                          PC( 2,ico) * recv1(ip+2,jp-2,kp-1))+ &
                            PC(-1,jco) * (PC(-2,ico) * recv1(ip-2,jp-1,kp-1) + &
                                          PC(-1,ico) * recv1(ip-1,jp-1,kp-1) + &
                                          PC( 0,ico) * recv1(ip,  jp-1,kp-1) + &
                                          PC( 1,ico) * recv1(ip+1,jp-1,kp-1) + &
                                          PC( 2,ico) * recv1(ip+2,jp-1,kp-1))+ &
                            PC( 0,jco) * (PC(-2,ico) * recv1(ip-2,jp,  kp-1) + &
                                          PC(-1,ico) * recv1(ip-1,jp,  kp-1) + &
                                          PC( 0,ico) * recv1(ip,  jp,  kp-1) + &
                                          PC( 1,ico) * recv1(ip+1,jp,  kp-1) + &
                                          PC( 2,ico) * recv1(ip+2,jp,  kp-1))+ &
                            PC( 1,jco) * (PC(-2,ico) * recv1(ip-2,jp+1,kp-1) + &
                                          PC(-1,ico) * recv1(ip-1,jp+1,kp-1) + &
                                          PC( 0,ico) * recv1(ip,  jp+1,kp-1) + &
                                          PC( 1,ico) * recv1(ip+1,jp+1,kp-1) + &
                                          PC( 2,ico) * recv1(ip+2,jp+1,kp-1))+ &
                            PC( 2,jco) * (PC(-2,ico) * recv1(ip-2,jp+2,kp-1) + &
                                          PC(-1,ico) * recv1(ip-1,jp+2,kp-1) + &
                                          PC( 0,ico) * recv1(ip,  jp+2,kp-1) + &
                                          PC( 1,ico) * recv1(ip+1,jp+2,kp-1) + &
                                          PC( 2,ico) * recv1(ip+2,jp+2,kp-1)))
            work(i,j,k,lb) = work(i,j,k,lb) + &
              PC( 0,kco) * (PC(-2,jco) * (PC(-2,ico) * recv1(ip-2,jp-2,kp  ) + &
                                          PC(-1,ico) * recv1(ip-1,jp-2,kp  ) + &
                                          PC( 0,ico) * recv1(ip,  jp-2,kp  ) + &
                                          PC( 1,ico) * recv1(ip+1,jp-2,kp  ) + &
                                          PC( 2,ico) * recv1(ip+2,jp-2,kp  ))+ &
                            PC(-1,jco) * (PC(-2,ico) * recv1(ip-2,jp-1,kp  ) + &
                                          PC(-1,ico) * recv1(ip-1,jp-1,kp  ) + &
                                          PC( 0,ico) * recv1(ip,  jp-1,kp  ) + &
                                          PC( 1,ico) * recv1(ip+1,jp-1,kp  ) + &
                                          PC( 2,ico) * recv1(ip+2,jp-1,kp  ))+ &
                            PC( 0,jco) * (PC(-2,ico) * recv1(ip-2,jp,  kp  ) + &
                                          PC(-1,ico) * recv1(ip-1,jp,  kp  ) + &
                                          PC( 0,ico) * recv1(ip,  jp,  kp  ) + &
                                          PC( 1,ico) * recv1(ip+1,jp,  kp  ) + &
                                          PC( 2,ico) * recv1(ip+2,jp,  kp  ))+ &
                            PC( 1,jco) * (PC(-2,ico) * recv1(ip-2,jp+1,kp  ) + &
                                          PC(-1,ico) * recv1(ip-1,jp+1,kp  ) + &
                                          PC( 0,ico) * recv1(ip,  jp+1,kp  ) + &
                                          PC( 1,ico) * recv1(ip+1,jp+1,kp  ) + &
                                          PC( 2,ico) * recv1(ip+2,jp+1,kp  ))+ &
                            PC( 2,jco) * (PC(-2,ico) * recv1(ip-2,jp+2,kp  ) + &
                                          PC(-1,ico) * recv1(ip-1,jp+2,kp  ) + &
                                          PC( 0,ico) * recv1(ip,  jp+2,kp  ) + &
                                          PC( 1,ico) * recv1(ip+1,jp+2,kp  ) + &
                                          PC( 2,ico) * recv1(ip+2,jp+2,kp  )))
            work(i,j,k,lb) = work(i,j,k,lb) + &
              PC( 1,kco) * (PC(-2,jco) * (PC(-2,ico) * recv1(ip-2,jp-2,kp+1) + &
                                          PC(-1,ico) * recv1(ip-1,jp-2,kp+1) + &
                                          PC( 0,ico) * recv1(ip,  jp-2,kp+1) + &
                                          PC( 1,ico) * recv1(ip+1,jp-2,kp+1) + &
                                          PC( 2,ico) * recv1(ip+2,jp-2,kp+1))+ &
                            PC(-1,jco) * (PC(-2,ico) * recv1(ip-2,jp-1,kp+1) + &
                                          PC(-1,ico) * recv1(ip-1,jp-1,kp+1) + &
                                          PC( 0,ico) * recv1(ip,  jp-1,kp+1) + &
                                          PC( 1,ico) * recv1(ip+1,jp-1,kp+1) + &
                                          PC( 2,ico) * recv1(ip+2,jp-1,kp+1))+ &
                            PC( 0,jco) * (PC(-2,ico) * recv1(ip-2,jp,  kp+1) + &
                                          PC(-1,ico) * recv1(ip-1,jp,  kp+1) + &
                                          PC( 0,ico) * recv1(ip,  jp,  kp+1) + &
                                          PC( 1,ico) * recv1(ip+1,jp,  kp+1) + &
                                          PC( 2,ico) * recv1(ip+2,jp,  kp+1))+ &
                            PC( 1,jco) * (PC(-2,ico) * recv1(ip-2,jp+1,kp+1) + &
                                          PC(-1,ico) * recv1(ip-1,jp+1,kp+1) + &
                                          PC( 0,ico) * recv1(ip,  jp+1,kp+1) + &
                                          PC( 1,ico) * recv1(ip+1,jp+1,kp+1) + &
                                          PC( 2,ico) * recv1(ip+2,jp+1,kp+1))+ &
                            PC( 2,jco) * (PC(-2,ico) * recv1(ip-2,jp+2,kp+1) + &
                                          PC(-1,ico) * recv1(ip-1,jp+2,kp+1) + &
                                          PC( 0,ico) * recv1(ip,  jp+2,kp+1) + &
                                          PC( 1,ico) * recv1(ip+1,jp+2,kp+1) + &
                                          PC( 2,ico) * recv1(ip+2,jp+2,kp+1)))
            work(i,j,k,lb) = work(i,j,k,lb) + &
              PC( 2,kco) * (PC(-2,jco) * (PC(-2,ico) * recv1(ip-2,jp-2,kp+2) + &
                                          PC(-1,ico) * recv1(ip-1,jp-2,kp+2) + &
                                          PC( 0,ico) * recv1(ip,  jp-2,kp+2) + &
                                          PC( 1,ico) * recv1(ip+1,jp-2,kp+2) + &
                                          PC( 2,ico) * recv1(ip+2,jp-2,kp+2))+ &
                            PC(-1,jco) * (PC(-2,ico) * recv1(ip-2,jp-1,kp+2) + &
                                          PC(-1,ico) * recv1(ip-1,jp-1,kp+2) + &
                                          PC( 0,ico) * recv1(ip,  jp-1,kp+2) + &
                                          PC( 1,ico) * recv1(ip+1,jp-1,kp+2) + &
                                          PC( 2,ico) * recv1(ip+2,jp-1,kp+2))+ &
                            PC( 0,jco) * (PC(-2,ico) * recv1(ip-2,jp,  kp+2) + &
                                          PC(-1,ico) * recv1(ip-1,jp,  kp+2) + &
                                          PC( 0,ico) * recv1(ip,  jp,  kp+2) + &
                                          PC( 1,ico) * recv1(ip+1,jp,  kp+2) + &
                                          PC( 2,ico) * recv1(ip+2,jp,  kp+2))+ &
                            PC( 1,jco) * (PC(-2,ico) * recv1(ip-2,jp+1,kp+2) + &
                                          PC(-1,ico) * recv1(ip-1,jp+1,kp+2) + &
                                          PC( 0,ico) * recv1(ip,  jp+1,kp+2) + &
                                          PC( 1,ico) * recv1(ip+1,jp+1,kp+2) + &
                                          PC( 2,ico) * recv1(ip+2,jp+1,kp+2))+ &
                            PC( 2,jco) * (PC(-2,ico) * recv1(ip-2,jp+2,kp+2) + &
                                          PC(-1,ico) * recv1(ip-1,jp+2,kp+2) + &
                                          PC( 0,ico) * recv1(ip,  jp+2,kp+2) + &
                                          PC( 1,ico) * recv1(ip+1,jp+2,kp+2) + &
                                          PC( 2,ico) * recv1(ip+2,jp+2,kp+2)))
#endif

          enddo
        enddo
      enddo

!===============================================================================

      if (monotone) then

! Monotonicity checking:  force interpolants to be flat if fine-zone averages
! fall outside the range of values of coarse-zone averages used to constrain
! them.

        icmin = ((ia-1)/2) + 1 + (nguard/2) + ioff
        icmax = ((ib-1)/2) + 1 + (nguard/2) + ioff
        jcmin = ((ja-1)/2) + 1 + (nguard/2)*k2d + joff
        jcmax = ((jb-1)/2) + 1 + (nguard/2)*k2d + joff
        kcmin = ((ka-1)/2) + 1 + (nguard/2)*k3d + koff
        kcmax = ((kb-1)/2) + 1 + (nguard/2)*k3d + koff

        k = ka - 2
        do kc = kcmin, kcmax
          k = k + 2
          j = ja - 2
          do jc = jcmin, jcmax
            j = j + 2
            i = ia - 2
            do ic = icmin, icmax
              i = i + 2

              cmin = minval(recv1(ic-1:ic+1,jc-k2d:jc+k2d, & 
     &                            kc-k3d:kc+k3d))
              cmax = maxval(recv1(ic-1:ic+1,jc-k2d:jc+k2d, & 
     &                            kc-k3d:kc+k3d))
              fmin = minval(work(i:i+1,j:j+k2d,k:k+k3d,lb))
              fmax = maxval(work(i:i+1,j:j+k2d,k:k+k3d,lb))

              if ( (fmin < cmin) .or. (fmax > cmax) ) then
                do kp = k, k+k3d
                  do jp = j, j+k2d
                    do ip = i, i+1
                      work(ip,jp,kp,lb) = recv1(ic,jc,kc)
                    enddo
                  enddo
                enddo
              endif

            enddo
          enddo
        enddo

      endif

!===============================================================================

      return
      end
